@@ -28,7 +28,7 @@
                 cl:*gensym-counter*
                 cl:*load-pathname*
                 cl:*load-print*
-                cl:*load-truename*
+                #+ansi-compliant-load-truename cl:*load-truename*
                 cl:*load-verbose*
                 cl:*macroexpand-hook*
                 cl:*modules*
@@ -65,14 +65,39 @@
                 cl:-
                 cl:/
                 cl://
-                cl:///)))
+                cl:///
+                ;; extension/internal specials are also proclaimed early
+                ;; to benefit from always-bound and precomputed TLS index.
+                sb-kernel:*current-level-in-print*
+                sb-ext:*print-vector-length*)))
     `(progn
        (declaim (special ,@list)
-                (sb!ext:always-bound ,@list))
+                (sb-ext:always-bound ,@list))
        (eval-when (:compile-toplevel :load-toplevel)
          (dolist (symbol ',list)
-           (declare (notinline (setf sb!int:info))) ; skirt failure-to-inline warning
-           (setf (sb!int:info :variable :wired-tls symbol) t)))))
+           (declare (notinline (setf sb-int:info))) ; skirt failure-to-inline warning
+           (setf (sb-int:info :variable :wired-tls symbol) t)))))
+
+#-ansi-compliant-load-truename
+(progn
+;; this could almost go in target-load.lisp were it not for the use in ir1-translators
+(defvar sb-fasl::%load-truename nil)
+;; "The consequences are unspecified if an attempt is made to assign or bind [...]"
+;;  http://www.lispworks.com/documentation/HyperSpec/Body/v_ld_pns.htm#STload-truenameST
+;;
+;; But the rationale provided in http://www.lispworks.com/documentation/HyperSpec/Issues/iss218_w.htm
+;; for the existence of the -TRUENAME* variables is dubious:
+;;  "Note that it is not adequate to just have the -PATHNAME* variables
+;;   since TRUENAME on these pathnames might not yield the value of the
+;;   -TRUENAME* variables if the file has been deleted or protected
+;;   since the open occurred (in some implementations)."
+;; because in our ANSI-compliant logic, there is a delay between opening a file
+;; and determining its truename, which means that it might be impossible to determine
+;; the truename even if done ASAP.  Supposing that somebody really were to delete
+;; the very file being acted on, it's not LOAD's job to deal with race conditions.
+(define-symbol-macro *load-truename*
+    (sb-ext:truly-the (values (or pathname null) &optional)
+      (sb-fasl::resolve-load-truename 'sb-fasl::%load-truename *load-pathname*))))
 
 (declaim (type t cl:+ cl:++ cl:+++ cl:- cl:* cl:** cl:***))
 
@@ -85,7 +110,7 @@
                          cl:*print-readably* cl:*read-eval*
                          cl:*read-suppress*))
 
-(declaim (type sb!pretty::pprint-dispatch-table cl:*print-pprint-dispatch*))
+(declaim (type sb-pretty:pprint-dispatch-table cl:*print-pprint-dispatch*))
 
 (declaim (type readtable cl:*readtable*))
 
@@ -93,12 +118,12 @@
 
 (declaim (type (member :upcase :downcase :capitalize) cl:*print-case*))
 
-(declaim (type (member cl:single-float cl:double-float
-                       cl:short-float cl:long-float) cl:*read-default-float-format*))
+(declaim (type (member single-float double-float short-float long-float cl:rational)
+               cl:*read-default-float-format*))
 
 (declaim (type list cl:/ cl:// cl:/// cl:*features* cl:*modules*))
 
-(declaim (type sb!kernel:type-specifier cl:*break-on-signals*))
+(declaim (type sb-kernel:type-specifier cl:*break-on-signals*))
 
 (declaim (type package cl:*package*))
 
@@ -117,12 +142,13 @@
                        cl:*query-io*
                        cl:*terminal-io*))
 
-(declaim (type (or function symbol) cl:*debugger-hook* cl:*macroexpand-hook*))
+(declaim (type sb-kernel:function-designator cl:*debugger-hook* cl:*macroexpand-hook*))
 
 (declaim (type unsigned-byte cl:*gensym-counter*))
 
 (declaim (type (or unsigned-byte null)
                        cl:*print-length*
+                       sb-ext:*print-vector-length*
                        cl:*print-level*
                        cl:*print-lines*
                        cl:*print-miser-width*
@@ -130,50 +156,27 @@
 
 (declaim (type pathname cl:*default-pathname-defaults*))
 
-(declaim (type (or pathname null)
-                       cl:*load-pathname*
-                       cl:*load-truename*
-                       cl:*compile-file-pathname*
-                       cl:*compile-file-truename*))
+;;; Some functions for which we should *never* (or almost never) bind eagerly
+;;; to the functional definition.
+;;; IR2-CONVERT-GLOBAL-VAR won't use a :KNOWN-FUN constant-tn, but will instead
+;;; dereference the fdefinition like for most function calls.
+;;; They aren't actually "inlined", but they were bypassing the fdefinition in
+;;; situations involving (APPLY ...) which rendered encapsulation impossible.
+(declaim (notinline open compile-file load compile))
 
-;;;; DEFGLOBAL and DEFINE-LOAD-TIME-GLOBAL
-;;;; These have alternate definitions (in cross-misc) which rely on
-;;;; the underlying host DEFVAR when building the cross-compiler.
+(in-package "SB-IMPL")
 
-(in-package "SB!IMPL")
 
-(defmacro defglobal (name value &optional (doc nil docp))
-  "Defines NAME as a global variable that is always bound. VALUE is evaluated
-and assigned to NAME both at compile- and load-time, but only if NAME is not
-already bound.
-
-Global variables share their values between all threads, and cannot be
-locally bound, declared special, defined as constants, and neither bound
-nor defined as symbol macros.
-
-See also the declarations SB-EXT:GLOBAL and SB-EXT:ALWAYS-BOUND."
-  (let ((boundp (make-symbol "BOUNDP")))
-    `(progn
-       (eval-when (:compile-toplevel)
-         (let ((,boundp (boundp ',name)))
-           (%compiler-defglobal ',name :always-bound
-                                (unless ,boundp ,value) (not ,boundp))))
-       (let ((,boundp (boundp ',name)))
-         (%defglobal ',name (unless ,boundp ,value) ,boundp ',doc ,docp
-                     (sb!c:source-location))))))
-
-(defmacro define-load-time-global (name value &optional (doc nil docp))
-  "Defines NAME as a global variable that is always bound. VALUE is evaluated
-and assigned to NAME at load-time, but only if NAME is not already bound.
-
-Attempts to read NAME at compile-time will signal an UNBOUND-VARIABLE error
-unless it has otherwise been assigned a value.
-
-See also DEFGLOBAL which assigns the VALUE at compile-time too."
-  (let ((boundp (make-symbol "BOUNDP")))
-    `(progn
-       (eval-when (:compile-toplevel)
-         (%compiler-defglobal ',name :eventually nil nil))
-       (let ((,boundp (boundp ',name)))
-         (%defglobal ',name (unless ,boundp ,value) ,boundp ',doc ,docp
-                     (sb!c:source-location))))))
+;;; Ensure some VM symbols get wired TLS.
+(in-package "SB-VM")
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (dolist (entry '#.per-thread-c-interface-symbols)
+    (let ((symbol (if (consp entry) (car entry) entry)))
+      (declare (notinline sb-int:info (setf sb-int:info)))
+      ;; CURRENT-{CATCH/UWP}-BLOCK are thread slots,
+      ;; so the TLS indices were already assigned.
+      ;; There may be other symbols too.
+      (unless (sb-int:info :variable :wired-tls symbol)
+        (setf (sb-int:info :variable :wired-tls symbol) :always-thread-local))
+      (unless (sb-int:info :variable :always-bound symbol)
+        (setf (sb-int:info :variable :always-bound symbol) :always-bound)))))
