@@ -41,6 +41,7 @@
                   *the-class-global-reader-method*
                   *the-class-global-writer-method*
                   *the-class-global-boundp-method*
+                  *the-class-global-makunbound-method*
                   *the-class-standard-generic-function*
                   *the-class-standard-direct-slot-definition*
                   *the-class-standard-effective-slot-definition*
@@ -975,7 +976,7 @@ between the ~A definition and the ~A definition"
 
 (define-type-method (classoid :negate) (type) (make-negation-type type))
 
-(define-type-method (classoid :unparse) (type)
+(define-type-method (classoid :unparse) (flags type)
   (classoid-proper-name type))
 
 ;;;; built-in classes
