@@ -25,7 +25,7 @@
 (defmacro finalizer-max-id (store) `(elt ,store 2))
 
 (defun make-finalizer-store (array-length)
-  (let* ((v (make-array (the index array-length)))
+  (let* ((v (make-array (the index array-length) :initial-element 0))
          (ht (make-system-hash-table :test 'eq :weakness :key :synchronized nil
                                      :finalizer t)))
     ;; The recycle bin has a dummy item in front so that the simple-vector
@@ -43,7 +43,7 @@
 (declaim (simple-vector **finalizer-store**))
 
 (defun finalize (object function &key dont-save
-                        &aux (item (if dont-save (list function) function)))
+                        &aux (function (%coerce-callable-to-fun function)))
   "Arrange for the designated FUNCTION to be called when there
 are no more references to OBJECT, including references in
 FUNCTION itself.
@@ -90,11 +90,22 @@ Examples:
     (finalize \"oops\" #'oops)
     (oops)) ; GC causes re-entry to #'oops due to the finalizer
             ; -> ERROR, caught, WARNING signalled"
-  (unless object
-    (error "Cannot finalize NIL."))
-  (with-finalizer-store (store)
-    (let ((id (gethash object (finalizer-id-map store))))
-      (cond (id ; object already has at least one finalizer
+  (declare (sb-c::tlab :system))
+  (let ((space (heap-allocated-p object)))
+    ;; Rule out immediate, stack, arena, readonly, and static objects.
+    ;; (Is it really an error for a readonly? Maybe a warning? I'll leave it this way unless
+    ;; users complain. Surely DX and arena are errors, and NIL was always an error.)
+    (unless (member space '(:dynamic :immobile))
+      (if (eq space :static)
+          (error "Cannot finalize ~S." object)
+          ;; silently discard finalizers on file streams in arenas I guess
+          (progn ; (warn "Will not finalize ~S." object)
+            (return-from finalize object)))))
+  (let ((item (if dont-save (list function) function)))
+    (with-finalizer-store (store)
+      (let ((id (gethash object (finalizer-id-map store))))
+        (cond
+            (id ; object already has at least one finalizer
              ;; Multiple finalizers are invoked in the order added.
              (let* ((old (svref store id))
                     (new (make-array (if (simple-vector-p old)
@@ -129,7 +140,7 @@ Examples:
              ;; Clear out lingering junk from (SVREF STORE ID) before
              ;; establishing that OBJECT maps to that index.
              (setf (svref store id) item
-                   (gethash object (finalizer-id-map store)) id)))))
+                   (gethash object (finalizer-id-map store)) id))))))
   object)
 
 (defun invalidate-fd-streams ()
@@ -158,7 +169,9 @@ Examples:
   ;; finalizers can in practice almost never be run, as pseudo-static objects
   ;; don't die, making this more-or-less an exercise in futility.
   (with-finalizer-store (old-store)
-    (without-gcing
+    ;; This doesn't need WITHOUT-GCING. MAPHASH will never present its funarg
+    ;; with a culled entry. GC during the MAPHASH could remove some items
+    ;; before we get to them, and that's fantastic.
       (let ((new-store
               (make-finalizer-store (max (1+ (finalizer-max-id old-store))
                                          +finalizers-initial-size+)))
@@ -181,7 +194,7 @@ Examples:
                  old-objects)
         (clrhash old-objects)
         (fill old-store 0)
-        (setq **finalizer-store** new-store)))))
+        (setq **finalizer-store** new-store))))
 
 ;;; Replace the finalizer store with a copy.  Tenured (gen6 = pseudo-static)
 ;;; vectors are problematic in many ways for gencgc, unless immutable.
@@ -227,27 +240,44 @@ Examples:
          ;; Not strictly necessary to do this: the next FINALIZE claiming
          ;; the same ID would assign a fresh list anyway.
          (setf (svref store it) 0)
-         (atomic-push it (finalizer-recycle-bin store)))))
-    object))
+         (locally (declare (sb-c::tlab :system))
+           (atomic-push it (finalizer-recycle-bin store)))))))
+  object)
 
 ;;; Drain the queue of finalizers and return when empty.
 ;;; Concurrent invocations of this function in different threads are ok.
 ;;; Nested invocations (from a GC forced by a finalizer) are not ok.
 ;;; See the trace at the bottom of this file.
 (defvar *in-a-finalizer* nil)
-(defun scan-finalizers ()
+#+sb-thread (define-alien-variable finalizer-thread-runflag int)
+
+(define-load-time-global *bg-compiler-function* nil)
+
+(defun run-pending-finalizers (&aux (hashtable (finalizer-id-map **finalizer-store**))
+                                    (ran-bg-compile)
+                                    (ran-a-system-finalizer)
+                                    (system-finalizer-scratchpad (list 0))
+                                    (ran-a-user-finalizer))
+  (declare (truly-dynamic-extent system-finalizer-scratchpad))
   ;; This never acquires the finalizer store lock. Code accordingly.
-  (let ((hashtable (finalizer-id-map **finalizer-store**)))
-    (loop
-     (let ((cell (hash-table-culled-values hashtable)))
-       ;; This is like atomic-pop, but its obtains the first cons cell
-       ;; in the list, not the car of the first cons.
-       ;; Possible TODO: when no other work remains, free the *JOINABLE-THREADS*,
-       ;; though MAKE-THREAD and JOIN-THREAD do that also, so there's no memory leak.
-       (loop (unless cell (return-from scan-finalizers))
-             (let ((actual (cas (hash-table-culled-values hashtable)
-                                cell (cdr cell))))
-               (if (eq actual cell) (return) (setq cell actual))))
+  (loop
+   ;; Perform no further work if trying to stop the thread, even if there is work.
+   #+sb-thread (when (zerop finalizer-thread-runflag) (return))
+   ;; Try to run a background compilation task
+   (when *bg-compiler-function* (setq ran-bg-compile (funcall *bg-compiler-function*)))
+   ;; Try to run 1 system finalizer
+   (setq ran-a-system-finalizer (sb-vm::immobile-code-dealloc-1 system-finalizer-scratchpad))
+   ;; Try to run 1 user finalizer
+   (let ((cell (hash-table-culled-values hashtable)))
+     ;; This is like atomic-pop, but its obtains the first cons cell
+     ;; in the list, not the car of the first cons.
+     ;; Possible TODO: when no other work remains, free the *JOINABLE-THREADS*,
+     ;; though MAKE-THREAD and JOIN-THREAD do that also, so there's no memory leak.
+     (loop (unless cell (return))
+           (let ((actual (cas (hash-table-culled-values hashtable)
+                              cell (cdr cell))))
+             (if (eq actual cell) (return) (setq cell actual))))
+     (when cell
        (let* ((id (the index (car cell)))
               ;; No other thread can modify **FINALIZER-STORE** at index ID
               ;; because the table no longer contains an object mapping to
@@ -274,6 +304,7 @@ Examples:
            (if (simple-vector-p finalizers)
                (map nil #'call finalizers)
                (call finalizers)))
+         (setq ran-a-user-finalizer t)
          ;; While the assignment to (SVREF STORE ID) should have been adequate,
          ;; we don't know that the vector is current - a new vector could have
          ;; gotten assigned into **FINALIZER-STORE** in between [1] and [2],
@@ -291,112 +322,63 @@ Examples:
          ;; removing dangling references, but if it's just a function,
          ;; there's nothing to smash.
          (cond ((simple-vector-p finalizers) (fill finalizers 0))
-               ((consp finalizers) (rplaca finalizers 0))))))))
+               ((consp finalizers) (rplaca finalizers 0))))))
+   ;; Did this iteration do anything at all?
+   (unless (or ran-bg-compile ran-a-system-finalizer ran-a-user-finalizer)
+     (return))
+   (setq ran-bg-compile nil
+         ran-a-system-finalizer nil
+         ran-a-user-finalizer nil)))
 
+(define-load-time-global *finalizer-thread* nil)
+(declaim (type (or sb-thread:thread (eql :start) null) *finalizer-thread*))
 #+sb-thread
 (progn
-  ;; *FINALIZER-THREAD* is either a boolean value indicating whether to start a
-  ;; thread, or a thread object (which is created no sooner than needed).
-  ;; Saving a core sets the flag to NIL so that finalizers which execute
-  ;; between stopping the thread and writing to disk will be synchronous.
-  ;; Restarting a saved core resets the flag to T.
-  (define-load-time-global *finalizer-thread* t)
-  (declaim (type (or sb-thread:thread boolean) *finalizer-thread*))
-  (define-load-time-global *finalizer-queue-lock*
-      (sb-thread:make-mutex :name "finalizer"))
-  (define-load-time-global *finalizer-queue*
-      (sb-thread:make-waitqueue :name "finalizer")))
+(defun finalizer-thread-notify ()
+  (alien-funcall (extern-alien "finalizer_thread_wake" (function void)))
+  nil)
 
-(defun run-pending-finalizers ()
-  (when (hash-table-culled-values (finalizer-id-map **finalizer-store**))
-    (cond #+sb-thread
-          ((%instancep *finalizer-thread*)
-           ;; Do absolutely nothing if in a finalizer. If we are then it's obvious
-           ;; that we're in the SCAN-FINALIZERS loop, and that the loop will just
-           ;; go on looping without further provocation.
-           (unless *in-a-finalizer*
-             (with-system-mutex (*finalizer-queue-lock*)
-               (sb-thread:condition-notify *finalizer-queue*))))
-          #+sb-thread
-          ((eq *finalizer-thread* t) ; Create a new thread
-           (sb-thread::make-ephemeral-thread
+;;; The following operations are synchronized by *MAKE-THREAD-LOCK* -
+;;;   FINALIZER-THREAD-{START,STOP}, S-L-A-D, SB-POSIX:FORK
+(defun finalizer-thread-start ()
+  (with-system-mutex (sb-thread::*make-thread-lock*)
+    #+(and unix sb-safepoint)
+    (sb-thread::make-system-thread "sigwait"
+                                   #'sb-unix::signal-handler-loop
+                                   nil 'sb-unix::*sighandler-thread*)
+    (aver (not *finalizer-thread*))
+    (setf finalizer-thread-runflag 1)
+    (setq *finalizer-thread* :start)
+    (let ((thread
+           (sb-thread::make-system-thread
             "finalizer"
             (lambda ()
-              ;; If we already called FINALIZER-THREAD-STOP, then
-              ;; *FINALIZER-THREAD* is NIL, and this WHEN test is false.
-              (when (eq t (cas *finalizer-thread* t
-                               sb-thread:*current-thread*))
-                ;; Don't enter the loop if this thread lost the
-                ;; competition to become a finalizer thread.
-                (loop
-                  (scan-finalizers)
-                  ;; Wait for a notification
-                  (with-system-mutex (*finalizer-queue-lock*)
-                    ;; Don't go to sleep if *FINALIZER-THREAD* became NIL
-                    (unless *finalizer-thread*
-                      (return))
-                    ;; The return value of CONDITION-WAIT is irrelevant
-                    ;; since it is always legal to call SCAN-FINALIZERS
-                    ;; even when it has nothing to do.
-                    ;; Spurious wakeup is of no concern to us here.
-                    (sb-thread:condition-wait
-                     *finalizer-queue* *finalizer-queue-lock*)))))
-            nil))
-          (t
-           (scan-finalizers)))))
+              (setf *finalizer-thread* sb-thread:*current-thread*)
+              (loop (run-pending-finalizers)
+                    (alien-funcall (extern-alien "finalizer_thread_wait" (function void)))
+                    (when (zerop finalizer-thread-runflag) (return)))
+              (setq *finalizer-thread* nil))
+            nil nil)))
+      ;; Don't return from this function until *FINALIZER-THREAD* has a good value,
+      ;; but don't set it to a thread if the thread was not created, or exited already.
+      (cas *finalizer-thread* :start thread))))
 
-;;; If a finalizer thread was started, stop it and wait for it to finish.
-;;; Make no attempt to drain the queue of pending finalizers.
-;;; (When called from EXIT, the user must invoke a final GC if there is
-;;; an expectation that GC-based things run. Similarly when saving a core)
-;;; There's a data race involved during SAVE-LISP-AND-DIE (or SB-POSIX:FORK)
-;;; because *FINALIZER-THREAD* is assigned by that thread itself which means that
-;;; it has to run before we can detect that it exists.
-;;; There are few remedies involving some combination of these steps:
-;;; - Assign a THREAD instance into *finalizer-thread* before it shows
-;;;   up in *ALL-THREADS*. If *FINALIZER-THREAD* has a value but is not in
-;;;   *ALL-THREADS*, then it may be blocked on the mutex. Can we terminate it?
-;;; - During save, acquire the *MAKE-THREAD-LOCK* to get a consistent view of
-;;;   which threads are starting and running.
-;;; - the :pauseless-threadstart code adds new threads to *ALL-THREADS*
-;;;   before starting the posix thread. We could look there.
-;;;   The important thing is to avoid thinking that there is a thread running
-;;;   when it doesn't have a posix thread ID yet.
-;;; - Does acquiring the *MAKE-THREAD-LOCK* before assigning *FINALIZER-THREAD*
-;;;   help anything? Perhaps.
-;;; But first I'd like to entirely eliminate the non-pauseless-threadstart code
-;;; so that we don't have to reason about two totally different solutions.
+;;; You should almost always invoke this with *MAKE-THREAD-LOCK* held.
+;;; Some tests violate that, but they know what they're doing.
 (defun finalizer-thread-stop ()
-  #+sb-thread
+  #+(and unix sb-safepoint)
+  (let ((thread sb-unix::*sighandler-thread*))
+    (aver (sb-thread::thread-p thread))
+    (setq sb-unix::*sighandler-thread* nil)
+    ;; This kill causes the thread's sigwait() syscall to return normally
+    ;; and then not invoke any handler.
+    (sb-unix:pthread-kill (sb-thread::thread-os-thread thread) sb-unix:sigterm)
+    (sb-thread:join-thread thread))
   (let ((thread *finalizer-thread*))
-    ;; valid state transitions:
-    ;;   T to a #<thread>
-    ;;   #<thread> to NIL
-    ;;   T to NIL
-    ;; If it was NIL, it is latched at that state, so we can ignore it.
-    (when (null thread)
-      (return-from finalizer-thread-stop))
-    ;; If user calls SAVE-LISP-AND-DIE or SB-POSIX:FORK in a finalizer...
-    (when (eq sb-thread:*current-thread* *finalizer-thread*)
-      (error "Can't stop finalizer thread from inside finalizer thread"))
-    (let ((oldval (cas *finalizer-thread* thread nil)))
-      ;; If GC had started the thread, and it got to its main body only after
-      ;; we observed *FINALIZER-THREAD* = T, then we could see a transition
-      ;; from T to #<thread> now. That is unlikely, but we must try once
-      ;; again to CAS the value to NIL.
-      (unless (eq oldval thread)
-        (setq thread oldval)
-        ;; Also, we could be racing with someone else trying to stop the
-        ;; thread, so we could see T | #<thread> -> NIL.
-        (when (null thread)
-          (return-from finalizer-thread-stop))
-        (aver (eq (cas *finalizer-thread* thread nil) thread))))
-    (when (%instancep thread)
-      ;; The finalizer thread will exit when we wake it up
-      ;; and it sees that *FINALIZER-THREAD* is NIL.
-      (with-system-mutex (*finalizer-queue-lock*)
-        (sb-thread:condition-notify *finalizer-queue*))
-      (sb-thread:join-thread thread)))) ; wait for it
+    (aver (sb-thread::thread-p thread))
+    (alien-funcall (extern-alien "finalizer_thread_stop" (function void)))
+    (sb-thread:join-thread thread)))
+)
 
 #|
 ;;; This is a display produced by annotating parts of gc-common.c and
