@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 ;;; Helper for making the DX closure allocation in macros expanding
 ;;; to CALL-WITH-FOO less ugly.
@@ -61,7 +61,7 @@
                           (stem (if (every #'alpha-char-p symbol-name)
                                     symbol-name
                                     (string (gensymify* symbol-name "-")))))
-                     `(,symbol (sb!xc:gensym ,stem))))
+                     `(,symbol (gensym ,stem))))
                  symbols)
      ,@body))
 
@@ -69,24 +69,22 @@
 ;;; macros and other code-manipulating code.)
 (defun make-gensym-list (n &optional name)
   (let ((arg (if name (string name) "G")))
-    (loop repeat n collect (sb!xc:gensym arg))))
+    (loop repeat n collect (gensym arg))))
 
 ;;;; miscellany
 
 ;;; Lots of code wants to get to the KEYWORD package or the
-;;; COMMON-LISP package without a lot of fuss, so we cache them in
-;;; variables on the host, or use L-T-V forms on the target.
-(macrolet ((def-it (sym expr)
-             #+sb-xc-host
-             `(progn (declaim (type package ,sym))
-                     (defvar ,sym ,expr))
-             #-sb-xc-host
-             ;; We don't need to declaim the type. FIND-PACKAGE
-             ;; returns a package, and L-T-V propagates types.
-             ;; It's ugly how it achieves that, but it's a separate concern.
-             `(define-symbol-macro ,sym (load-time-value ,expr t))))
-  (def-it *cl-package* (find-package "COMMON-LISP"))
-  (def-it *keyword-package* (find-package "KEYWORD")))
+;;; COMMON-LISP package without going through FIND-PACKAGE, so we refer to them
+;;; as constants which is ever so slightly more efficient than a defglobal.
+;;; DEFINE-SYMBOL-MACRO should be ok in any host lisp. We used to distrust it,
+;;; but it is specified by CLHS, as are package constants, so there should
+;;; be no need to fear this idiom.
+(macrolet ((def-it (sym name) `(define-symbol-macro ,sym ,(find-package name))))
+  ;; *CL-PACKAGE* is always COMMON-LISP, not XC-STRICT-CL on the host, because the latter
+  ;; is just a means to avoid inheriting symbols that are not supposed to be in the CL:
+  ;; package but might be due to non-ansi-compliance of the host.
+  (def-it *cl-package* "COMMON-LISP")
+  (def-it *keyword-package* "KEYWORD"))
 
 (declaim (inline singleton-p))
 (defun singleton-p (list)
@@ -94,9 +92,10 @@
 
 (defun gensymify (x)
   (if (symbolp x)
-      (sb!xc:gensym (symbol-name x))
-      (sb!xc:gensym)))
+      (gensym (symbol-name x))
+      (gensym)))
 
+(eval-when (:load-toplevel :execute #+sb-xc-host :compile-toplevel)
 (labels ((symbol-concat (package &rest things)
            (dx-let ((strings (make-array (length things)))
                     (length 0)
@@ -108,7 +107,7 @@
                              (l (length s)))
                         (setf (svref strings index) s)
                         (incf length l)
-                        #!+sb-unicode
+                        #+sb-unicode
                         (when (and (typep s '(array character (*)))
                                    ;; BASE-CHAR-p isn't a standard predicate.
                                    ;; and host ignores ELT-TYPE anyway.
@@ -124,13 +123,13 @@
                  (setq name (make-array length :element-type elt-type)))
                (dotimes (index (length things)
                                (if package
-                                   (values (%intern name length package elt-type))
+                                   (values (%intern name length package elt-type nil))
                                    (make-symbol name)))
                  (let ((s (svref strings index)))
                    (replace name s :start1 start)
                    (incf start (length s)))))))
-         #+sb-xc-host (%intern (name length package elt-type)
-                        (declare (ignore length elt-type))
+         #+sb-xc-host (%intern (name length package elt-type dummy)
+                        (declare (ignore length elt-type dummy))
                         ;; Copy, in case the host respects the DX declaration,
                         ;; but does not copy, which makes our assumption wrong.
                         (intern (copy-seq name) package)))
@@ -140,7 +139,7 @@
     (apply #'symbol-concat (sane-package) things))
   ;; SYMBOLICATE in given package.
   (defun package-symbolicate (package &rest things)
-    (apply #'symbol-concat package things))
+    (apply #'symbol-concat (find-package package) things))
   ;; like SYMBOLICATE, but producing keywords
   (defun keywordicate (&rest things)
     (apply #'symbol-concat *keyword-package* things))
@@ -149,7 +148,7 @@
   ;; consistency with the above would not be particularly enlightening
   ;; as to how it isn't just GENSYMIFY]
   (defun gensymify* (&rest things)
-    (apply #'symbol-concat nil things)))
+    (apply #'symbol-concat nil things))))
 
 ;;; Access *PACKAGE* in a way which lets us recover when someone has
 ;;; done something silly like (SETF *PACKAGE* :CL-USER) in unsafe code.
@@ -179,8 +178,7 @@
           (t
            ;; We're in the undefined behavior zone. First, munge the
            ;; system back into a defined state.
-           (let ((really-package
-                  (load-time-value (find-package :cl-user) t)))
+           (let ((really-package #.(find-package :cl-user)))
              (setf *package* really-package)
              ;; Then complain.
              (error 'simple-type-error
@@ -247,6 +245,13 @@
 (defun ensure-list (thing)
   (if (listp thing) thing (list thing)))
 
+(defun recons (old-cons car cdr)
+  "If CAR is eq to the car of OLD-CONS and CDR is eq to the CDR, return
+  OLD-CONS, otherwise return (cons CAR CDR)."
+  (if (and (eq car (car old-cons)) (eq cdr (cdr old-cons)))
+      old-cons
+      (cons car cdr)))
+
 ;;; Anaphoric macros
 (defmacro awhen (test &body body)
   `(let ((it ,test))
@@ -265,15 +270,18 @@
                       it)
                  (acond ,@rest)))))))
 
-;; This is not an 'extension', but is needed super early, so ....
-(defmacro sb!xc:defconstant (name value &optional (doc nil docp))
-  "Define a global constant, saying that the value is constant and may be
-  compiled into code. If the variable already has a value, and this is not
-  EQL to the new value, the code is not portable (undefined behavior). The
-  third argument is an optional documentation string for the variable."
-  `(eval-when (:compile-toplevel :load-toplevel :execute)
-     (sb!c::%defconstant ',name ,value (sb!c:source-location)
-                         ,@(and docp `(',doc)))))
+;;; Like GETHASH if HASH-TABLE contains an entry for KEY.
+;;; Otherwise, evaluate DEFAULT, store the resulting value in
+;;; HASH-TABLE and return two values: 1) the result of evaluating
+;;; DEFAULT 2) NIL.
+(defmacro ensure-gethash (key hash-table default)
+  (with-unique-names (n-key n-hash-table value foundp)
+    `(let ((,n-key ,key)
+           (,n-hash-table ,hash-table))
+       (multiple-value-bind (,value ,foundp) (gethash ,n-key ,n-hash-table)
+         (if ,foundp
+             (values ,value t)
+             (values (setf (gethash ,n-key ,n-hash-table) ,default) nil))))))
 
 (defvar *!removable-symbols* nil)
 
@@ -294,39 +302,15 @@
 ;;;
 #+sb-xc-host
 (defmacro defconstant-eqx (symbol expr eqx &optional doc)
-  ;; CLISP needs the EVAL-WHEN here, or else the symbol defined is unavailable
-  ;; for later uses within the same file. For instance, in x86-64/vm, defining
-  ;; TEMP-REG-TN as R11-TN would get an error that R11-TN is unbound.
-  ;; We don't want that junk in our expansion, especially as our requirement
-  ;; is to separate the compile-time and load-time effect.
   `(eval-when (:compile-toplevel :load-toplevel :execute)
-     (defconstant ,symbol (%defconstant-eqx-value ',symbol ,expr ,eqx)
+     ;; We use DEFVAR rather than DEFCONSTANT as a host effect in
+     ;; order to avoid differences in host decisions about inlining
+     ;; the value of the constant, with knock-on effects on EQLity
+     ;; of references to internal parts of the constant.
+     (defvar ,symbol (%defconstant-eqx-value ',symbol ,expr ,eqx)
        ,@(when doc (list doc)))))
 
-;;; This is the expansion as it should be for us, but notice that this
-;;; recapitulation of the macro is actually not defined at compile-time.
 #-sb-xc-host
-(let ()
-  (defmacro defconstant-eqx (symbol expr eqx &optional doc)
-    `(defconstant ,symbol (%defconstant-eqx-value ',symbol ,expr ,eqx)
-       ,@(when doc (list doc)))))
-
-;;; Special variant at cross-compile-time. Face it: the "croak-if-not-EQx" test
-;;; is irrelevant - there can be no pre-existing value to test against.
-;;; The extra magic is that we need to discern between constants simple enough
-;;; to assigned during genesis (cold-load) from those assigned in cold-init.
-;;; This choice informs the compiler how to emit references to the symbol.
-(defvar sb!c::*!const-value-deferred* '())
-#-sb-xc-host
-(eval-when (:compile-toplevel)
-  (sb!xc:defmacro defconstant-eqx (symbol expr eqx &optional doc)
-    (let ((constp (sb!xc:constantp expr)))
-      `(progn
-         (eval-when (:compile-toplevel)
-           (sb!xc:defconstant ,symbol (%defconstant-eqx-value ',symbol ,expr ,eqx))
-           ,@(unless constp
-               `((push ',symbol sb!c::*!const-value-deferred*))))
-         (eval-when (:load-toplevel)
-           (sb!c::%defconstant ',symbol
-             ,(if constp `',(constant-form-value expr) expr)
-             (sb!c:source-location) ,@(when doc (list doc))))))))
+(defmacro defconstant-eqx (symbol expr eqx &optional doc)
+  `(defconstant ,symbol (%defconstant-eqx-value ',symbol ,expr ,eqx)
+     ,@(when doc (list doc))))
