@@ -16,10 +16,27 @@
 (defvar *load-source-default-type* "lisp"
   "The source file types which LOAD looks for by default.")
 
+(defvar *load-verbose* nil
+  ;; Note that CMU CL's default for this was T, and ANSI says it's
+  ;; implementation-dependent. We choose NIL on the theory that it's
+  ;; a nicer default behavior for Unix programs.
+  "the default for the :VERBOSE argument to LOAD")
+
+(defvar *load-print* nil
+  "the default for the :PRINT argument to LOAD")
+
+#+ansi-compliant-load-truename
 (defvar *load-truename* nil
   "the TRUENAME of the file that LOAD is currently loading")
+#-ansi-compliant-load-truename
+(setf (documentation '*load-truename* 'variable)
+  "the TRUENAME of the file that LOAD is currently loading")
+
 (defvar *load-pathname* nil
   "the defaulted pathname that LOAD is currently loading")
+
+(declaim (type (or pathname null) *load-truename* *load-pathname*))
+
 
 ;;;; LOAD-AS-SOURCE
 
@@ -119,10 +136,9 @@
                             (error () nil)))
          ;; FIXME: this ANSI-specified nonsense should have been an accessor call
          ;; because eager binding might force dozens of syscalls to occur.
-         ;; And you wonder why I/O is slow.  What a joke. Making this a symbol-macro
-         ;; that calls a function would be technically incompatible, but we could
-         ;; make it an unbound symbol and trap the access, stuffing in a value
-         ;; just-in-time. But BOUNDP would also need to hacked to return T.
+         ;; The non-ansi-compliant code resolves *LOAD-TRUENAME* only if referenced.
+         #-ansi-compliant-load-truename (%load-truename (make-unbound-marker))
+         #+ansi-compliant-load-truename
          (*load-truename* (when *load-pathname*
                             (handler-case (truename stream)
                               (file-error () nil))))
@@ -130,6 +146,13 @@
          (*load-depth* (1+ *load-depth*)))
     (funcall function stream arg)))
 
+(defun resolve-load-truename (symbol untruename)
+  (if (boundp symbol)
+      (symbol-value symbol)
+      (set symbol (when untruename
+                    (handler-case (truename untruename)
+                      (file-error () nil))))))
+
 ;;; Returns T if the stream is a binary input stream with a FASL header.
 (defun fasl-header-p (stream &key errorp)
   (unless (and (member (stream-element-type stream) '(character base-char))
@@ -171,10 +194,26 @@
                              :expected header))))
           (file-position stream p))))))
 
-(defun load (pathspec &key (verbose *load-verbose*) (print *load-print*)
-                           (if-does-not-exist t) (external-format :default))
-  "Load the file given by FILESPEC into the Lisp environment, returning
-   T on success."
+(defun load (filespec &key (verbose *load-verbose*) (print *load-print*)
+                           (if-does-not-exist :error) (external-format :default))
+  "Load the file given by FILESPEC into the Lisp environment, returning T on
+   success. The file type (a.k.a extension) is defaulted if missing. These
+   options are defined:
+
+   :IF-DOES-NOT-EXIST
+       If :ERROR (the default), signal an error if the file can't be located.
+       If NIL, simply return NIL (LOAD normally returns T.)
+
+   :VERBOSE
+       If true, print a line describing each file loaded.
+
+   :PRINT
+       If true, print information about loaded values.  When loading the
+       source, the result of evaluating each top-level form is printed.
+
+   :EXTERNAL-FORMAT
+       The external-format to use when opening the FILENAME. The default is
+       :DEFAULT which uses the SB-EXT:*DEFAULT-EXTERNAL-FORMAT*."
   (labels ((load-stream (stream faslp)
              (if (and (fd-stream-p stream)
                       (eq (sb-impl::fd-stream-fd-type stream) :directory))
@@ -186,11 +225,11 @@
                   #'load-stream-1 stream
                   faslp
                   ;; If you prefer *LOAD-PATHNAME* to reflect what the user specified prior
-                  ;; to merging, then CALL-WITH-LOAD-BINDINGS is passed PATHSPEC,
+                  ;; to merging, then CALL-WITH-LOAD-BINDINGS is passed FILESPEC,
                   ;; otherwise it is passed STREAM.
                   (cond ((and (not sb-c::*merge-pathnames*)
-                              (typep pathspec '(or string pathname)))
-                         pathspec)
+                              (typep filespec '(or string pathname)))
+                         filespec)
                         (t stream)))))
            (load-stream-1 (stream faslp)
              (let (;; Bindings required by ANSI.
@@ -207,7 +246,7 @@
                    (sb-c::*policy* sb-c::*policy*)
                    (sb-c::*handled-conditions* sb-c::*handled-conditions*))
                (if faslp
-                   (load-as-fasl stream verbose print)
+                   (load-as-fasl stream verbose (if print t nil))
                    ;; FIXME: if *EVALUATOR-MODE* is :INTERPRET,
                    ;; then this should have nothing whatsoever to do with
                    ;; compiler-error-resignaling. That's an artifact
@@ -217,17 +256,17 @@
     (declare (truly-dynamic-extent #'load-stream-1))
 
     ;; Case 1: stream.
-    (when (streamp pathspec)
-      (return-from load (load-stream pathspec (fasl-header-p pathspec))))
+    (when (streamp filespec)
+      (return-from load (load-stream filespec (fasl-header-p filespec))))
 
-    (let ((pathname (pathname pathspec)))
+    (let ((pathname (pathname filespec)))
       ;; Case 2: Open as binary, try to process as a fasl.
       (with-open-stream
-          (stream (or (open pathspec :element-type '(unsigned-byte 8)
+          (stream (or (open filespec :element-type '(unsigned-byte 8)
                                      :if-does-not-exist nil)
-                      (when (null (pathname-type pathspec))
+                      (when (null (pathname-type filespec))
                         (let ((defaulted-pathname
-                                (probe-load-defaults pathspec)))
+                                (probe-load-defaults filespec)))
                           (if defaulted-pathname
                               (progn (setq pathname defaulted-pathname)
                                      (open pathname
@@ -236,11 +275,28 @@
                                            :element-type '(unsigned-byte 8))))))
                       (if if-does-not-exist
                           (error 'simple-file-error
-                                 :pathname pathspec
+                                 :pathname filespec
                                  :format-control
                                  "~@<Couldn't load ~S: file does not exist.~@:>"
-                                 :format-arguments (list pathspec))
+                                 :format-arguments (list filespec))
                           (return-from load nil))))
+        ;; This is the understandable logic.
+        #-ansi-compliant-load-truename
+        (if (string-equal (pathname-type pathname) *fasl-file-type*)
+            (load-stream stream t)
+            (with-open-file (stream pathname :external-format external-format
+                                             :class 'form-tracking-stream)
+              (load-stream stream nil)))
+        ;; I have no idea what "don't allow empty .fasls" was supposed to mean.
+        ;; The more natural interpretation to me would be that if a file is empty,
+        ;; we don't treat the file as fasl, but instead treat it as source (and do nothing)
+        ;; regardless of the name. But the actual effect seems to be to force an error.
+        ;; So I would have commented that as "force an error on empty fasls", especially
+        ;; as there is ambiguity in "don't X and Y" - does it mean "don't X" and "don't Y"?
+        ;; Because surely that would be illogical - "don't assume empty files are source".
+        ;; Anyway, reasonable users should never test the edge cases of this,
+        ;; and so all it effectively achieves is slowing down the loading of files.
+        #+ansi-compliant-load-truename
         (let* ((real (probe-file stream))
                (should-be-fasl-p
                  (and real (string-equal (pathname-type real) *fasl-file-type*))))
@@ -295,11 +351,6 @@
 
 ;;;; linkage fixups
 
-;;; Lisp assembler routines are named by Lisp symbols, not strings,
-;;; and so can be compared by EQ.
-(define-load-time-global *assembler-routines* nil)
-(declaim (code-component *assembler-routines*))
-
 (defun calc-asm-routine-bounds ()
   (loop for v being each hash-value of (%code-debug-info *assembler-routines*)
         minimize (car v) into min
@@ -312,13 +363,27 @@
 (defvar *!initial-assembler-routines*)
 
 (defun get-asm-routine (name &optional indirect &aux (code *assembler-routines*))
+  ;; Each architecture can define an "indirect" value in its own peculiar way.
+  ;; Only some of our architectures use that option.
   (awhen (the list (gethash (the symbol name) (%code-debug-info code)))
-    (sap-int (sap+ (code-instructions code)
-                   (if indirect
-                       ;; Return the address containing the routine address
-                       (ash (cddr it) sb-vm:word-shift)
-                       ;; Return the routine address itself
-                       (car it))))))
+    (destructuring-bind (start end . index) it
+      (declare (ignore end) (ignorable index))
+      (let* ((insts (code-instructions code))
+             (addr (sap-int (sap+ insts start))))
+        (unless indirect
+          (return-from get-asm-routine addr))
+        #-(or ppc ppc64 x86 x86-64) (bug "Indirect asm-routine lookup")
+        #+(or ppc ppc64) (- addr sb-vm:nil-value)
+        #+(or x86 x86-64) ; return the address of a word containing 'addr'
+        (let ((offset (ash index sb-vm:word-shift)))
+          ;; the address is in the "external" static-space jump table
+          #+immobile-space (+ (get-lisp-obj-address *asm-routine-vector*)
+                              (ash sb-vm:vector-data-offset sb-vm:word-shift)
+                          ;; offset is biased by 1 word, accounting for the jump-table-count
+                          ;; at the first word in code-instructions. So unbias it.
+                              (- offset sb-vm:n-word-bytes sb-vm:other-pointer-lowtag))
+          #-immobile-space ; the address is in the "internal" jump table
+          (sap-int (sap+ insts offset)))))))
 
 (defun !loader-cold-init ()
   (let* ((code *assembler-routines*)
@@ -327,10 +392,13 @@
          (count (length vector))
          (ht (make-hash-table))) ; keys are symbols
     (rplaca (%code-debug-info code) ht)
+    (setf *asm-routine-index-to-name* (make-array (1+ (length vector))))
     (dotimes (i count)
       (destructuring-bind (name . offset) (svref vector i)
         (let ((next-offset (if (< (1+ i) count) (cdr (svref vector (1+ i))) size)))
-          (aver (> next-offset offset))
+          (setf (aref *asm-routine-index-to-name* (1+ i)) name)
+          ;; Must be in ascending order, but one address can have more than one name.
+          (aver (>= next-offset offset))
           ;; store inclusive bounds on PC offset range and the function index
           (setf (gethash name ht) (list* offset (1- next-offset) (1+ i))))))))
 
@@ -341,9 +409,3 @@
     (abort-build ()
       :report "Abort building SBCL."
       (sb-ext:exit :code 1))))
-
-;;; Remember where cold artifacts went, and put the warm ones there too
-;;; because it looks nicer not to scatter them throughout the source tree.
-;;; *t-o-prefix* isn't known to the compiler, and we need it to be
-;;; initialized from a constant, so use read-time eval.
-(defvar *!target-obj-prefix* #.sb-cold::*target-obj-prefix*)
