@@ -69,7 +69,9 @@
     (error "Heap underflow"))
   (prog1
       (aref heap i)
-    (setf (aref heap i) (aref heap (1- (length heap))))
+    (let ((last (1- (length heap))))
+      (setf (aref heap i) (aref heap last)
+            (aref heap last) 0))
     (decf (fill-pointer heap))
     (heapify heap i :key key :test test)))
 
@@ -188,7 +190,7 @@ from now. For timers with a repeat interval it returns true."
 
 (defmacro with-scheduler-lock ((&optional) &body body)
   ;; Don't let the SIGALRM handler mess things up.
-  `(sb-thread::with-system-mutex (*scheduler-lock*)
+  `(with-system-mutex (*scheduler-lock*)
      ,@body))
 
 (defun under-scheduler-lock-p ()
@@ -344,13 +346,10 @@ triggers."
 ;;; win32 waitable timers using a timerfd-like portability layer in
 ;;; the runtime.
 
-#+sb-wtimer
-(define-alien-type wtimer
-    #+win32 system-area-pointer ;HANDLE, but that's not defined yet
-    #+sunos system-area-pointer ;struct os_wtimer *
-    #+(or android linux bsd) int)
+#+win32
+(define-alien-type wtimer system-area-pointer) ;HANDLE, but that's not defined yet
 
-#+sb-wtimer
+#+win32
 (progn
   (define-alien-routine "os_create_wtimer" wtimer)
   (define-alien-routine "os_wait_for_wtimer" int (wt wtimer))
@@ -370,7 +369,7 @@ triggers."
         (prog1
             (setf *waitable-timer-handle* (os-create-wtimer))
           (setf *timer-thread*
-                (sb-thread::make-ephemeral-thread
+                (sb-thread::make-system-thread
                  "System timer watchdog thread"
                  (lambda ()
                    (loop while
@@ -378,7 +377,7 @@ triggers."
                               (os-wait-for-wtimer *waitable-timer-handle*))
                              *waitable-timer-handle*)
                          doing (run-expired-timers)))
-                 nil)))))
+                 nil nil)))))
 
   (defun itimer-emulation-deinit ()
     (with-scheduler-lock ()
@@ -411,7 +410,7 @@ triggers."
               (values 0 min-nsec)
               (values s u))))))
 
-#-(or sb-wtimer win32)
+#+unix
 (progn
   (defun %set-system-timer (sec nsec)
     (sb-unix:unix-setitimer :real 0 0 sec (ceiling nsec 1000)))
