@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;;; DEFKNOWNs
 
@@ -66,8 +66,16 @@
 (defknown (setf %alien-value) (t system-area-pointer unsigned-byte alien-type) t
   ())
 
-(defknown alien-funcall (alien-value &rest *) *
+(defknown alien-funcall (alien-value &rest t) *
   (any recursive))
+
+(defknown sb-alien::string-to-c-string (simple-string t) (or (simple-array (unsigned-byte 8) (*))
+                                                             simple-base-string)
+    (movable flushable))
+(defknown sb-alien::c-string-to-string (system-area-pointer t t) simple-string
+    (movable flushable))
+(defknown sb-alien::c-string-external-format * *
+        (movable flushable))
 
 ;;;; cosmetic transforms
 
@@ -155,7 +163,7 @@
   (multiple-value-bind (slot-offset slot-type)
       (find-slot-offset-and-type alien slot)
     (/noshow "in DEFTRANSFORM %SLOT-ADDR, creating %SAP-ALIEN")
-    `(%sap-alien (sap+ (alien-sap alien) (/ ,slot-offset sb!vm:n-byte-bits))
+    `(%sap-alien (sap+ (alien-sap alien) (/ ,slot-offset sb-vm:n-byte-bits))
                  ',(make-alien-pointer-type :to slot-type))))
 
 ;;;; DEREF support
@@ -164,10 +172,7 @@
   (let ((alien-type (lvar-type alien)))
     (unless (alien-type-type-p alien-type)
       (give-up-ir1-transform))
-    (let ((alien-type (alien-type-type-alien-type alien-type)))
-      (if (alien-type-p alien-type)
-          alien-type
-          (give-up-ir1-transform)))))
+    (alien-type-type-alien-type alien-type)))
 
 (defun find-deref-element-type (alien)
   (let ((alien-type (find-deref-alien-type alien)))
@@ -231,7 +236,6 @@
 
 #+nil ;; Shouldn't be necessary.
 (defoptimizer (deref derive-type) ((alien &rest noise))
-  (declare (ignore noise))
   (block nil
     (catch 'give-up-ir1-transform
       (return (make-alien-type-type (find-deref-element-type alien))))
@@ -247,7 +251,6 @@
 
 #+nil ;; ### Again, the value might be coerced.
 (defoptimizer (%set-deref derive-type) ((alien value &rest noise))
-  (declare (ignore noise))
   (block nil
     (catch 'give-up-ir1-transform
       (let ((type (make-alien-type-type
@@ -267,7 +270,6 @@
              value))))
 
 (defoptimizer (%deref-addr derive-type) ((alien &rest noise))
-  (declare (ignore noise))
   (block nil
     (catch 'give-up-ir1-transform
       (return (make-alien-type-type
@@ -280,7 +282,7 @@
       (compute-deref-guts alien indices)
     (/noshow "in DEFTRANSFORM %DEREF-ADDR, creating (LAMBDA .. %SAP-ALIEN)")
     `(lambda (alien ,@indices-args)
-       (%sap-alien (sap+ (alien-sap alien) (/ ,offset-expr sb!vm:n-byte-bits))
+       (%sap-alien (sap+ (alien-sap alien) (/ ,offset-expr sb-vm:n-byte-bits))
                    ',(make-alien-pointer-type :to element-type)))))
 
 ;;;; support for aliens on the heap
@@ -316,7 +318,7 @@
           (return type))))
     *wild-type*))
 
-(deftransform %set-heap-alien ((info value) (heap-alien-info *) *)
+(deftransform %set-heap-alien ((info value) (heap-alien-info t) *)
   (multiple-value-bind (sap type) (heap-alien-sap-and-type info)
     `(setf (%alien-value ,sap 0 ',type) value)))
 
@@ -351,22 +353,22 @@
     (/noshow (local-alien-info-force-to-memory-p info))
     (/noshow alien-type (unparse-alien-type alien-type) (alien-type-bits alien-type))
     (if (local-alien-info-force-to-memory-p info)
-        #!+(or x86 x86-64)
+        #+(or x86 x86-64)
         `(%primitive alloc-alien-stack-space
                      ,(ceiling (alien-type-bits alien-type)
-                               sb!vm:n-byte-bits))
-        #!-(or x86 x86-64)
+                               sb-vm:n-byte-bits))
+        #-(or x86 x86-64)
         `(%primitive alloc-number-stack-space
                      ,(ceiling (alien-type-bits alien-type)
-                               sb!vm:n-byte-bits))
+                               sb-vm:n-byte-bits))
         (let* ((alien-rep-type-spec (compute-alien-rep-type alien-type))
                (alien-rep-type (specifier-type alien-rep-type-spec)))
           (cond ((csubtypep (specifier-type 'system-area-pointer)
                             alien-rep-type)
                  '(int-sap 0))
                 ((ctypep 0 alien-rep-type) 0)
-                ((ctypep 0.0f0 alien-rep-type) 0.0f0)
-                ((ctypep 0.0d0 alien-rep-type) 0.0d0)
+                ((ctypep $0.0f0 alien-rep-type) $0.0f0)
+                ((ctypep $0.0d0 alien-rep-type) $0.0d0)
                 (t
                  (compiler-error
                   "Aliens of type ~S cannot be represented immediately."
@@ -409,7 +411,6 @@
         '(error "This should be eliminated as dead code."))))
 
 (defoptimizer (%local-alien-addr derive-type) ((info var))
-  (declare (ignore var))
   (if (constant-lvar-p info)
       (let* ((info (lvar-value info))
              (alien-type (local-alien-info-type info)))
@@ -428,7 +429,6 @@
 ;;;; %CAST
 
 (defoptimizer (%cast derive-type) ((alien type))
-  (declare (ignore alien))
   (or (when (constant-lvar-p type)
         (let ((alien-type (lvar-value type)))
           (when (alien-type-p alien-type)
@@ -461,7 +461,6 @@
        (give-up-ir1-transform)))))
 
 (defoptimizer (%sap-alien derive-type) ((sap type))
-  (declare (ignore sap))
   (if (constant-lvar-p type)
       (make-alien-type-type (lvar-value type))
       *wild-type*))
@@ -495,113 +494,15 @@
   (deftransform (setf %alien-value) ((value sap offset type))
     (%computed-lambda #'compute-deposit-lambda type)))
 
-;;;; a hack to clean up divisions
-
-(defun count-low-order-zeros (thing)
-  (typecase thing
-    (lvar
-     (if (constant-lvar-p thing)
-         (count-low-order-zeros (lvar-value thing))
-         (count-low-order-zeros (lvar-uses thing))))
-    (combination
-     (case (let ((name (lvar-fun-name (combination-fun thing))))
-             (or (modular-version-info name :untagged nil) name))
-       ((+ -)
-        (let ((min most-positive-fixnum)
-              (itype (specifier-type 'integer)))
-          (dolist (arg (combination-args thing) min)
-            (if (csubtypep (lvar-type arg) itype)
-                (setf min (min min (count-low-order-zeros arg)))
-                (return 0)))))
-       (*
-        (let ((result 0)
-              (itype (specifier-type 'integer)))
-          (dolist (arg (combination-args thing) result)
-            (if (csubtypep (lvar-type arg) itype)
-                (setf result (+ result (count-low-order-zeros arg)))
-                (return 0)))))
-       (ash
-        (let ((args (combination-args thing)))
-          (if (= (length args) 2)
-              (let ((amount (second args)))
-                (if (constant-lvar-p amount)
-                    (max (+ (count-low-order-zeros (first args))
-                            (lvar-value amount))
-                         0)
-                    0))
-              0)))
-       (t
-        0)))
-    (integer
-     (if (zerop thing)
-         most-positive-fixnum
-         (do ((result 0 (1+ result))
-              (num thing (ash num -1)))
-             ((logbitp 0 num) result))))
-    (cast
-     (count-low-order-zeros (cast-value thing)))
-    (t
-     0)))
-
-(deftransform / ((numerator denominator) (integer integer))
-  "convert x/2^k to shift"
-  (unless (constant-lvar-p denominator)
-    (give-up-ir1-transform))
-  (let* ((denominator (lvar-value denominator))
-         (bits (1- (integer-length denominator))))
-    (unless (and (> denominator 0) (= (ash 1 bits) denominator))
-      (give-up-ir1-transform))
-    (let ((alignment (count-low-order-zeros numerator)))
-      (unless (>= alignment bits)
-        (give-up-ir1-transform))
-      `(ash numerator ,(- bits)))))
-
-(deftransform ash ((value amount))
-  (let ((value-node (lvar-uses value)))
-    (unless (combination-p value-node)
-      (give-up-ir1-transform))
-    (let ((inside-fun-name (lvar-fun-name (combination-fun value-node))))
-      (multiple-value-bind (prototype width)
-          (modular-version-info inside-fun-name :untagged nil)
-        (unless (eq (or prototype inside-fun-name) 'ash)
-          (give-up-ir1-transform))
-        (when (and width (not (constant-lvar-p amount)))
-          (give-up-ir1-transform))
-        (let ((inside-args (combination-args value-node)))
-          (unless (= (length inside-args) 2)
-            (give-up-ir1-transform))
-          (let ((inside-amount (second inside-args)))
-            (unless (and (constant-lvar-p inside-amount)
-                         (not (minusp (lvar-value inside-amount))))
-              (give-up-ir1-transform)))
-          (splice-fun-args value inside-fun-name 2)
-          (if width
-              `(lambda (value amount1 amount2)
-                 (logand (ash value (+ amount1 amount2))
-                         ,(1- (ash 1 (+ width (lvar-value amount))))))
-              `(lambda (value amount1 amount2)
-                 (ash value (+ amount1 amount2)))))))))
-
 ;;;; ALIEN-FUNCALL support
 
 (deftransform alien-funcall ((function &rest args)
-                             ((alien (* t)) &rest *) *)
+                             ((alien (* t)) &rest t) *)
   (let ((names (make-gensym-list (length args))))
     (/noshow "entering first DEFTRANSFORM ALIEN-FUNCALL" function args)
     `(lambda (function ,@names)
        (alien-funcall (deref function) ,@names))))
 
-#-sb-xc-host
-(defun find-saved-fp-and-pc (fp)
-  (dolist (x *saved-fp-and-pcs*)
-    (declare (type (simple-array word (2)) x))
-    (when (#!+stack-grows-downward-not-upward
-           sap>
-           #!-stack-grows-downward-not-upward
-           sap<
-           (int-sap (aref x 0)) fp)
-      (return (values (int-sap (aref x 0)) (int-sap (aref x 1)))))))
-
 (deftransform alien-funcall ((function &rest args) * * :node node)
   (let ((type (lvar-type function)))
     (unless (alien-type-type-p type)
@@ -677,9 +578,9 @@
             ;; Remember this frame to make sure that we can get back
             ;; to it later regardless of how the foreign stack looks
             ;; like.
-            #!+c-stack-is-control-stack
+            #+c-stack-is-control-stack
             (when (policy node (= 3 alien-funcall-saves-fp-and-pc))
-              (setf body `(invoke-with-saved-fp-and-pc (lambda () ,body))))
+              (setf body `(invoke-with-saved-fp (lambda () ,body))))
             (/noshow "returning from DEFTRANSFORM ALIEN-FUNCALL" (params) body)
             `(lambda (function ,@(params))
                ,@(when ignore-fun '((declare (ignore function))))
@@ -687,19 +588,16 @@
                ,body)))))))
 
 (defoptimizer (%alien-funcall derive-type) ((function type &rest args))
-  (declare (ignore function args))
   (unless (and (constant-lvar-p type)
                (alien-fun-type-p (lvar-value type)))
     (error "Something is broken."))
-  (let ((type (lvar-value type)))
-    (values-specifier-type
-     (compute-alien-rep-type
-      (alien-fun-type-result-type type)
-      :result))))
+  (let ((spec (compute-alien-rep-type
+               (alien-fun-type-result-type (lvar-value type))
+               :result)))
+    (if (eq spec '*) *wild-type* (values-specifier-type spec))))
 
 (defoptimizer (%alien-funcall ltn-annotate)
-              ((function type &rest args) node ltn-policy)
-  (declare (ignore type ltn-policy))
+              ((function type &rest args) node)
   (setf (basic-combination-info node) :funny)
   (setf (node-tail-p node) nil)
   (unless (and (constant-lvar-p function)
@@ -726,106 +624,123 @@
         ;; in reverse order here doesn't change the semantics, but we
         ;; deal with all of the stack arguments before the wired
         ;; register arguments become live.
-        (args #!-arm args #!+arm (reverse args))
-        #!+c-stack-is-control-stack
+        (args #-arm args #+arm (reverse args))
+        #+c-stack-is-control-stack
         (stack-pointer (make-stack-pointer-tn)))
     (multiple-value-bind (nsp stack-frame-size arg-tns result-tns)
         (make-call-out-tns type)
-      #!+x86
+      #+x86
       (vop set-fpu-word-for-c call block)
       ;; Save the stack pointer, it will get aligned and subtracting
       ;; the size will not restore the original value, and some
       ;; things, like SB-C::CALL-VARIABLE, use the stack pointer to
       ;; calculate the number of saved values.
       ;; See alien.impure.lisp/:stack-misalignment
-      #!+c-stack-is-control-stack
+      #+c-stack-is-control-stack
       (vop current-stack-pointer call block stack-pointer)
       (vop alloc-number-stack-space call block stack-frame-size nsp)
       ;; KLUDGE: This is where the second half of the ARM
       ;; register-pressure change lives (see above).
-      (dolist (tn #!-arm arg-tns #!+arm (reverse arg-tns))
-        ;; On PPC, TN might be a list. This is used to indicate
-        ;; something special needs to happen. See below.
-        ;;
-        ;; FIXME: We should implement something better than this.
-        (let* ((first-tn (if (listp tn) (car tn) tn))
-               (arg (pop args))
-               (sc (tn-sc first-tn))
-               (scn (sc-number sc))
-               (move-arg-vops (svref (sc-move-arg-vops sc) scn)))
-          (aver arg)
-          (unless (= (length move-arg-vops) 1)
-            (error "no unique move-arg-vop for moves in SC ~S" (sc-name sc)))
-          #!+(or x86 x86-64) (emit-move-arg-template call
-                                                     block
-                                                     (first move-arg-vops)
-                                                     (lvar-tn call block arg)
-                                                     nsp
-                                                     first-tn)
-          #!-(or x86 x86-64)
-          (cond
-            #!+arm-softfp
-            ((and (proper-list-of-length-p tn 3)
-                  (symbolp (third tn)))
-             (emit-template call block
-                            (template-or-lose (third tn))
-                            (reference-tn (lvar-tn call block arg) nil)
-                            (reference-tn-list (butlast tn) t)))
-            (t
-             (let ((temp-tn (make-representation-tn
-                             (tn-primitive-type first-tn) scn)))
-               (emit-move call
-                          block
-                          (lvar-tn call block arg)
-                          temp-tn)
-               (emit-move-arg-template call
-                                       block
-                                       (first move-arg-vops)
-                                       temp-tn
-                                       nsp
-                                       first-tn))))
-          #!+(and ppc darwin)
-          (when (listp tn)
-            ;; This means that we have a float arg that we need to
-            ;; also copy to some int regs. The list contains the TN
-            ;; for the float as well as the TNs to use for the int
-            ;; arg.
-            (destructuring-bind (float-tn i1-tn &optional i2-tn)
-                tn
-              (if i2-tn
-                  (vop sb!vm::move-double-to-int-arg call block
-                       float-tn i1-tn i2-tn)
-                  (vop sb!vm::move-single-to-int-arg call block
-                       float-tn i1-tn))))))
+      (dolist (tn #-arm arg-tns #+arm (reverse arg-tns))
+        (if (functionp tn)
+            (funcall tn (pop args) call block nsp)
+            ;; On PPC, TN might be a list. This is used to indicate
+            ;; something special needs to happen. See below.
+            ;;
+            ;; FIXME: We should implement something better than this.
+            (let* ((first-tn (if (listp tn) (car tn) tn))
+                   (arg (pop args))
+                   (sc (tn-sc first-tn))
+                   (scn (sc-number sc))
+                   (move-arg-vops (svref (sc-move-arg-vops sc) scn)))
+              (aver arg)
+              (unless (= (length move-arg-vops) 1)
+                (error "no unique move-arg-vop for moves in SC ~S" (sc-name sc)))
+
+              (cond
+                #+arm-softfp
+                ((and (listp tn)
+                      (symbolp (car (last tn))))
+                 (emit-template call block
+                                (template-or-lose (car (last tn)))
+                                (reference-tn (lvar-tn call block arg) nil)
+                                (reference-tn-list (butlast tn) t)))
+                (t
+                 (when (eq (sb-kind (sc-sb sc)) :unbounded) ;; stacks are unbounded
+                   ;; Avoid allocating this TN on the caller's stack
+                   (setf (tn-kind first-tn) :arg-pass))
+                 #+(or x86 x86-64)
+                 (emit-move-arg-template call
+                                         block
+                                         (first move-arg-vops)
+                                         (lvar-tn call block arg)
+                                         nsp
+                                         first-tn)
+                 #-(or x86 x86-64)
+                 (let* ((primitive-type (tn-primitive-type first-tn))
+                        ;; If the destination is a stack TN make sure
+                        ;; the temporary TN is a register.
+                        (scn (if (sc-number-stack-p sc)
+                                 (car (primitive-type-scs primitive-type))
+                                 scn))
+                        (temp-tn (make-representation-tn primitive-type scn)))
+                   (emit-move call block (lvar-tn call block arg) temp-tn)
+                   (emit-move-arg-template call
+                                           block
+                                           (first move-arg-vops)
+                                           temp-tn
+                                           nsp
+                                           first-tn))))
+              #+(and ppc darwin)
+              (when (listp tn)
+                ;; This means that we have a float arg that we need to
+                ;; also copy to some int regs. The list contains the TN
+                ;; for the float as well as the TNs to use for the int
+                ;; arg.
+                (destructuring-bind (float-tn i1-tn &optional i2-tn)
+                    tn
+                  (if i2-tn
+                      (vop sb-vm::move-double-to-int-arg call block
+                           float-tn i1-tn i2-tn)
+                      (vop sb-vm::move-single-to-int-arg call block
+                           float-tn i1-tn)))))))
       (aver (null args))
       (let* ((result-tns (ensure-list result-tns))
              (arg-operands
               (reference-tn-list (remove-if-not #'tn-p (flatten-list arg-tns)) nil))
              (result-operands
               (reference-tn-list (remove-if-not #'tn-p result-tns) t)))
-        (cond #!+(vop-named sb!vm::call-out-named)
+        (cond #+#.(cl:if (sb-c::vop-existsp :named sb-vm::call-out-named) '(and) '(or))
               ((and (constant-lvar-p function) (stringp (lvar-value function)))
                (vop* call-out-named call block (arg-operands) (result-operands)
                      (lvar-value function)
-                     (sb!alien::alien-fun-type-varargs type)))
+                     (sb-alien::alien-fun-type-varargs type)))
               (t
                (vop* call-out call block
                      ((lvar-tn call block function) arg-operands)
                      (result-operands))))
-        #!-c-stack-is-control-stack
+        #-c-stack-is-control-stack
         (vop dealloc-number-stack-space call block stack-frame-size)
-        #!+c-stack-is-control-stack
+        #+c-stack-is-control-stack
         (vop reset-stack-pointer call block stack-pointer)
-        #!+x86
+        #+x86
         (vop set-fpu-word-for-lisp call block)
         (cond
-          #!+arm-softfp
+          #+arm-softfp
           ((and lvar
-                (proper-list-of-length-p result-tns 3)
-                (symbolp (third result-tns)))
+                (symbolp (car (last result-tns))))
            (emit-template call block
-                          (template-or-lose (third result-tns))
-                          (reference-tn-list (butlast result-tns) nil)
-                          (reference-tn (car (ir2-lvar-locs (lvar-info lvar))) t)))
+                          (template-or-lose (car (last result-tns)))
+                          (reference-tn-list (butlast result-tns 2) nil)
+                          (reference-tn (car (last result-tns 2)) t))
+           (move-lvar-result call block (list (car (last result-tns 2))) lvar))
           (t
            (move-lvar-result call block result-tns lvar)))))))
+
+(deftransform sb-alien::c-string-external-format ((type)
+                                                  ((constant-arg sb-alien::alien-c-string-type)))
+  (let ((format (sb-alien::alien-c-string-type-external-format
+                 (lvar-value type))))
+    (if (eq format :default)
+        `(sb-alien::default-c-string-external-format)
+        `',format)))
