@@ -428,9 +428,7 @@
     ;; sanity checks
     (aver (zerop (hash-table-count (fasl-output-patch-table fasl-output))))
     ;; End the group.
-    (dump-fop 'fop-verify-empty-stack fasl-output)
-    (dump-fop 'fop-verify-table-size fasl-output (fasl-output-table-free fasl-output))
-    (dump-fop 'fop-end-group fasl-output))
+    (dump-fop 'fop-end-group fasl-output (fasl-output-table-free fasl-output)))
 
   ;; That's all, folks.
   (close (fasl-output-stream fasl-output) :abort abort-p)
@@ -1033,10 +1031,12 @@
   (assert (<= (length +fixup-kinds+) 8))) ; fixup-kind fits in 3 bits
 
 (defconstant-eqx +fixup-flavors+
-  #(:assembly-routine :assembly-routine* :asm-routine-nil-offset
+  #(:assembly-routine :assembly-routine*
     :gc-barrier :symbol-tls-index
-    :foreign :foreign-dataref :code-object
-    :layout :immobile-symbol :named-call :static-call
+    :alien-code-linkage-index :alien-data-linkage-index
+    :foreign :foreign-dataref
+    :code-object
+    :layout :immobile-symbol :fdefn-call :static-call
     :symbol-value
     :layout-id)
   #'equalp)
@@ -1109,7 +1109,7 @@
                           (t name)))))
               (:layout-id
                (the wrapper name))
-              ((:assembly-routine :assembly-routine* :asm-routine-nil-offset
+              ((:assembly-routine :assembly-routine*
                :symbol-tls-index
                ;; Only #+immobile-space can use the following two flavors.
                ;; An :IMMOBILE-SYMBOL fixup references the symbol itself,
@@ -1118,8 +1118,9 @@
                ;; but its global value must be an immobile object.
                :immobile-symbol :symbol-value)
                (the symbol name))
-              ((:foreign :foreign-dataref) (the string name))
-              ((:named-call :static-call) name))))
+              ((:alien-code-linkage-index :alien-data-linkage-index
+                :foreign :foreign-dataref) (the string name))
+              ((:fdefn-call :static-call) name))))
       (dump-object info fasl-output)
       (incf nelements (cond (named (dump-object operand fasl-output) 2)
                             (t 1))))))
@@ -1143,7 +1144,7 @@
   (let* ((2comp (component-info component))
          (constants (sb-c:ir2-component-constants 2comp))
          (header-length (length constants))
-         (n-named-calls 0))
+         (n-fdefns 0))
     (collect ((patches)
               (named-constants))
       ;; Dump the constants, noting any :ENTRY constants that have to
@@ -1175,8 +1176,12 @@
                     (dump-fop 'fop-misc-trap fasl-output)))))
                (:load-time-value
                 (dump-push (cadr entry) fasl-output))
-               ((:named-call :fdefinition)
-                (when (eq (car entry) :named-call) (incf n-named-calls))
+               (:fdefinition
+                ;; It's possible for other fdefns to be found in the header, but they can't
+                ;; have resulted from IR2 conversion. They would have had to come from
+                ;; something like (load-time-value (find-or-create-fdefn ...))
+                ;; which is fine, but they don't count for this purpose.
+                (incf n-fdefns)
                 (dump-object (cadr entry) fasl-output)
                 (dump-fop 'fop-fdefn fasl-output))
                (:known-fun
@@ -1209,7 +1214,7 @@
       (dump-integer-as-n-bytes (length (sb-c::ir2-component-entries 2comp))
                                4 ; output 4 bytes
                                fasl-output)
-      (dump-integer-as-n-bytes (the (unsigned-byte 22) n-named-calls)
+      (dump-integer-as-n-bytes (the (unsigned-byte 22) n-fdefns)
                                4 ; output 4 bytes
                                fasl-output)
       (dump-segment code-segment code-length fasl-output)
@@ -1261,8 +1266,6 @@
   (declare (type component component))
   (declare (type fasl-output file))
 
-  (dump-fop 'fop-verify-table-size file (fasl-output-table-free file))
-
   #+sb-dyncount
   (let ((info (sb-c::ir2-component-dyncount-info (component-info component))))
     (when info
