@@ -15,10 +15,10 @@
 
 ;;;; coercions
 
-(deftransform float ((n f) (* single-float) *)
+(deftransform float ((n f) (t single-float) *)
   '(%single-float n))
 
-(deftransform float ((n f) (* double-float) *)
+(deftransform float ((n f) (t double-float) *)
   '(%double-float n))
 
 (deftransform float ((n) *)
@@ -29,13 +29,25 @@
 (deftransform %single-float ((n) (single-float) *)
   'n)
 
+(deftransform %single-float ((n) (ratio) *)
+  '(sb-kernel::float-ratio n 'single-float))
+
+(deftransform %single-float ((n) (bignum) *)
+  '(bignum-to-float n 'single-float))
+
 (deftransform %double-float ((n) (double-float) *)
   'n)
 
+(deftransform %double-float ((n) (ratio) *)
+  '(sb-kernel::float-ratio n 'double-float))
+
+(deftransform %double-float ((n) (bignum) *)
+  '(bignum-to-float n 'double-float))
+
 ;;; RANDOM
 (macrolet ((frob (fun type)
              `(deftransform random ((num &optional state)
-                                    (,type &optional *) *)
+                                    (,type &optional t) *)
                 "Use inline float operations."
                 '(,fun num (or state *random-state*)))))
   (frob %random-single-float single-float)
@@ -67,7 +79,7 @@
 ;;; a division to get the random value into the desired range.
 (deftransform random ((num &optional state)
                       ((constant-arg (integer 1 #.(expt 2 sb-vm:n-word-bits)))
-                       &optional *)
+                       &optional t)
                       *
                       :policy (and (> speed compilation-speed)
                                    (> speed space)))
@@ -220,7 +232,7 @@
 (deftransform integer-decode-float ((x) (double-float) *)
   '(integer-decode-double-float x))
 
-(deftransform scale-float ((f ex) (single-float *) *)
+(deftransform scale-float ((f ex) (single-float t) *)
   (cond #+x86
         ((csubtypep (lvar-type ex)
                     (specifier-type '(signed-byte 32)))
@@ -228,7 +240,7 @@
         (t
          '(scale-single-float f ex))))
 
-(deftransform scale-float ((f ex) (double-float *) *)
+(deftransform scale-float ((f ex) (double-float t) *)
   (cond #+x86
         ((csubtypep (lvar-type ex)
                     (specifier-type '(signed-byte 32)))
@@ -322,6 +334,17 @@
         most-negative-single-float most-positive-single-float)
   (frob %double-float double-float
         most-negative-double-float most-positive-double-float))
+
+(defoptimizer (float derive-type) ((number prototype))
+  (let ((type (lvar-type prototype)))
+    (unless (or (csubtypep type (specifier-type 'double-float))
+                (csubtypep type (specifier-type 'single-float)))
+      (handler-case
+          (type-union
+           (one-arg-derive-type number #'%single-float-derive-type-aux #'%single-float)
+           (one-arg-derive-type number #'%double-float-derive-type-aux #'%double-float))
+        (type-error ()
+          nil)))))
 
 ;;;; float contagion
 
@@ -569,7 +592,7 @@
            'single-float))
 (deftransform expt ((x y) (double-float double-float) double-float)
   `(%pow x y))
-(deftransform expt ((x y) (single-float integer) double-float)
+(deftransform expt ((x y) (single-float integer) single-float)
   `(coerce (%pow (coerce x 'double-float) (coerce y 'double-float))
     'single-float))
 (deftransform expt ((x y) (double-float integer) double-float)
@@ -844,7 +867,7 @@
        (list (interval-expt-> x y-)
              (interval-expt-> x y+))))))
 
-;;; Handle the case when x <= 1
+;;; Handle the case when 0 <= x <= 1
 (defun interval-expt-< (x y)
   (case (interval-range-info x $0d0)
     (+
@@ -875,9 +898,13 @@
                 (interval-expt-< x y+))))))
     (-
      ;; The case where x <= 0. Y MUST be an INTEGER for this to work!
-     ;; The calling function must insure this! For now we'll just
-     ;; return the appropriate unbounded float type.
-     (list (make-interval :low nil :high nil)))
+     ;; The calling function must insure this!
+     (loop for interval in (flatten-list (interval-expt (interval-neg x) y))
+           for low = (interval-low interval)
+           for high = (interval-high interval)
+           collect interval
+           when (or high low)
+           collect (interval-neg interval)))
     (t
      (destructuring-bind (neg pos)
          (interval-split 0 x t t)
@@ -1526,22 +1553,106 @@
 
 
 ;;;; TRUNCATE, FLOOR, CEILING, and ROUND
+(deftransform truncate ((x &optional by)
+                        (t &optional (constant-arg (member 1))))
+  '(unary-truncate x))
 
-(macrolet ((define-frobs (fun ufun)
-             `(deftransform ,fun ((x &optional by)
-                                  (* &optional (constant-arg (member 1))))
-                  '(let ((res (,ufun x)))
-                    (values res (locally
-                                    (declare (flushable %single-float
-                                                        %double-float))
-                                  (- x res)))))))
-  (define-frobs truncate %unary-truncate)
-  (define-frobs round %unary-round))
+(deftransform round ((x &optional by)
+                     (t &optional (constant-arg (member 1))))
+  '(let ((res (%unary-round x)))
+    (values res (locally
+                    (declare (flushable %single-float
+                                        %double-float))
+                  (- x res)))))
 
 (deftransform %unary-truncate ((x) (single-float))
-  `(%unary-truncate/single-float x))
+  `(values (unary-truncate x)))
 (deftransform %unary-truncate ((x) (double-float))
-  `(%unary-truncate/double-float x))
+  `(values (unary-truncate x)))
+
+(defun value-within-numeric-type (type)
+  (labels ((try (x)
+             (when (ctypep x type)
+               (return-from value-within-numeric-type x)))
+           #-sb-xc-host
+           (next-float (float)
+             (multiple-value-bind (frac exp sign)
+                 (integer-decode-float float)
+               (* (scale-float (float (1+ frac) float) exp)
+                  sign)))
+           #-sb-xc-host
+           (prev-float (float)
+             (multiple-value-bind (frac exp sign)
+                 (integer-decode-float float)
+               (* (scale-float (float (1- frac) float) exp)
+                  sign)))
+           (next (x)
+             (typecase x
+               (integer
+                (1+ x))
+               #-sb-xc-host
+               (float
+                (next-float x))
+               (t
+                0)))
+           (prev (x)
+             (typecase x
+               (integer
+                (1- x))
+               #-sb-xc-host
+               (float
+                (prev-float x))
+               (t
+                0)))
+           (ratio-between (low high)
+             (+ low (/ (- high low) 2)))
+           (numeric (x)
+             (when (numeric-type-p x)
+               (let ((lo (numeric-type-low x))
+                     (hi (numeric-type-high x)))
+                 (when (numberp lo)
+                   (try lo))
+                 (when (numberp hi)
+                   (try hi))
+                 (when (consp lo)
+                   (try (next (car lo))))
+                 (when (consp hi)
+                   (try (prev (car hi))))
+                 (when (and (typep lo '(cons rational))
+                            (typep hi '(cons rational)))
+                   (try (ratio-between (car lo) (car hi))))
+                 (when (csubtypep x (specifier-type 'rational))
+                   (try 0))
+                 (when (csubtypep x (specifier-type 'double-float))
+                   (try $0d0))
+                 (when (csubtypep x (specifier-type 'single-float))
+                   (try $0f0))))))
+    (typecase type
+      (numeric-type (numeric type))
+      (union-type (mapc #'numeric (union-type-types type))))
+    (error "Couldn't come up with a value for ~s" type)))
+
+(deftransform unary-truncate ((x) * * :result result :node node)
+  (unless (lvar-single-value-p result)
+    (give-up-ir1-transform))
+  (let ((rem-type (second (values-type-required (node-derived-type node)))))
+    `(values (%unary-truncate x)
+             ,(value-within-numeric-type rem-type))))
+
+(macrolet ((def (type)
+             `(deftransform unary-truncate ((number) (,type))
+                '(if (typep number
+                      '(,type
+                        ,(symbol-value (package-symbolicate :sb-kernel 'most-negative-fixnum- type))
+                        ,(symbol-value (package-symbolicate :sb-kernel 'most-positive-fixnum- type))))
+                  (let ((truncated (truly-the fixnum (,(symbolicate '%unary-truncate/ type) number))))
+                    (declare (flushable ,(symbolicate "%" type)))
+                    (values truncated
+                            (- number
+                               (coerce truncated ',type))))
+                  (,(symbolicate 'unary-truncate- type '-to-bignum) number)))))
+  (def single-float)
+  (def double-float))
 
 ;;; Convert (TRUNCATE x y) to the obvious implementation.
 ;;;
@@ -1555,8 +1666,16 @@
 ;;; depending on FLOAT-ACCURACY. Finally, leave out the secondary value when
 ;;; we know it is unused: COERCE is not flushable.
 (macrolet ((def (type other-float-arg-types)
-             (let ((unary (symbolicate "%UNARY-TRUNCATE/" type))
-                   (coerce (symbolicate "%" type)))
+             (let* ((unary (symbolicate "%UNARY-TRUNCATE/" type))
+                    (unary-to-bignum (symbolicate '%unary-truncate- type '-to-bignum))
+                    (coerce (symbolicate "%" type))
+                    (unary `(lambda (number)
+                              (if (typep number
+                                         '(,type
+                                           ,(symbol-value (package-symbolicate :sb-kernel 'most-negative-fixnum- type))
+                                           ,(symbol-value (package-symbolicate :sb-kernel 'most-positive-fixnum- type))))
+                                  (truly-the fixnum (,unary number))
+                                  (,unary-to-bignum number)))))
                `(deftransform truncate ((x &optional y)
                                         (,type
                                          &optional (or ,type ,@other-float-arg-types integer))
@@ -1569,20 +1688,25 @@
                     (if (or (not y)
                             (and (constant-lvar-p y) (= 1 (lvar-value y))))
                         (if compute-all
-                            `(let ((res (,',unary x)))
-                               (values res (- x (locally
-                                                    ;; Can be flushed as it will produce no errors.
-                                                    (declare (flushable ,',coerce))
-                                                    (,',coerce res)))))
+                            `(unary-truncate x)
                             `(let ((res (,',unary x)))
                                ;; Dummy secondary value!
                                (values res x)))
                         (if compute-all
                             `(let* ((f (,',coerce y))
-                                    (res (,',unary (/ x f))))
-                               (values res (- x (* f (locally
-                                                         (declare (flushable ,',coerce))
-                                                       (,',coerce res))))))
+                                    (div (/ x f))
+                                    (res (,',unary div)))
+                               (values res
+                                       (- x (* f
+                                               #+round-float
+                                               (,',(ecase type
+                                                     (double-float 'round-double)
+                                                     (single-float 'round-single))
+                                                div :truncate)
+                                               #-round-float
+                                               (locally
+                                                   (declare (flushable ,',coerce))
+                                                 (,',coerce res))))))
                             `(let* ((f (,',coerce y))
                                     (res (,',unary (/ x f))))
                                ;; Dummy secondary value!
@@ -1673,7 +1797,7 @@
     (values (complex re im)
             (locally (declare (notinline complex)) (complex re im)))))
 
-#+(and (not sb-xc-host) (not riscv)) ; broken on riscv
+#-sb-xc-host
 (dolist (test '(try-folding-complex-single try-folding-complex-double))
   (multiple-value-bind (a b) (funcall test)
     (assert (eql a b)))
@@ -1692,3 +1816,23 @@
           (complex single-float-negative-infinity single-float-positive-infinity)
           (complex single-float-negative-infinity single-float-negative-infinity)
           (complex single-float-positive-infinity single-float-negative-infinity)))
+
+#+round-float
+(deftransform fround ((number &optional divisor) (double-float t))
+  (if (or (not divisor)
+          (and (constant-lvar-p divisor)
+               (= (lvar-value divisor) 1)))
+      `(let ((res (round-double number :round)))
+         (values res (- number res)))
+      `(let ((res (round-double (/ number (%double-float divisor)) :round)))
+         (values res (- number (* res divisor))))))
+
+#+round-float
+(deftransform fround ((number &optional divisor) (single-float (or null single-float rational)))
+  (if (or (not divisor)
+          (and (constant-lvar-p divisor)
+               (= (lvar-value divisor) 1)))
+      `(let ((res (round-single number :round)))
+         (values res (- number res)))
+      `(let ((res (round-single (/ number (%single-float divisor)) :round)))
+         (values res (- number (* res divisor))))))
