@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; primitive type definitions
 
@@ -18,33 +18,32 @@
 
 (!def-primitive-type t (descriptor-reg))
 (/show0 "primtype.lisp 20")
-(setf *backend-t-primitive-type* (primitive-type-or-lose t))
 
 ;;; primitive integer types that fit in registers
 (/show0 "primtype.lisp 24")
 (!def-primitive-type positive-fixnum (any-reg signed-reg unsigned-reg)
-  :type (unsigned-byte #.sb!vm:n-positive-fixnum-bits))
+  :type (unsigned-byte #.sb-vm:n-positive-fixnum-bits))
 (/show0 "primtype.lisp 27")
-#!-64-bit-registers
+#-(or 64-bit 64-bit-registers)
 (!def-primitive-type unsigned-byte-31 (signed-reg unsigned-reg descriptor-reg)
   :type (unsigned-byte 31))
 (/show0 "primtype.lisp 31")
-#!-64-bit-registers
+#-(or 64-bit 64-bit-registers)
 (!def-primitive-type unsigned-byte-32 (unsigned-reg descriptor-reg)
   :type (unsigned-byte 32))
 (/show0 "primtype.lisp 35")
-#!+64-bit-registers
+#+(or 64-bit 64-bit-registers)
 (!def-primitive-type unsigned-byte-63 (signed-reg unsigned-reg descriptor-reg)
   :type (unsigned-byte 63))
-#!+64-bit-registers
+#+(or 64-bit 64-bit-registers)
 (!def-primitive-type unsigned-byte-64 (unsigned-reg descriptor-reg)
   :type (unsigned-byte 64))
 (!def-primitive-type fixnum (any-reg signed-reg)
   :type (signed-byte #.(1+ n-positive-fixnum-bits)))
-#!-64-bit-registers
+#-(or 64-bit 64-bit-registers)
 (!def-primitive-type signed-byte-32 (signed-reg descriptor-reg)
   :type (signed-byte 32))
-#!+64-bit-registers
+#+(or 64-bit 64-bit-registers)
 (!def-primitive-type signed-byte-64 (signed-reg descriptor-reg)
   :type (signed-byte 64))
 
@@ -52,16 +51,17 @@
 
 (/show0 "primtype.lisp 53")
 (!def-primitive-type-alias tagged-num '(:or positive-fixnum fixnum))
-(multiple-value-bind (unsigned signed)
-    (case sb!vm::n-machine-word-bits
+(multiple-value-bind (unsigned signed untagged)
+    (case sb-vm:n-machine-word-bits
       (64 (values '(unsigned-byte-64 unsigned-byte-63 positive-fixnum)
-                  '(signed-byte-64 fixnum unsigned-byte-63 positive-fixnum)))
+                  '(signed-byte-64 fixnum unsigned-byte-63 positive-fixnum)
+                  '(signed-byte-64 unsigned-byte-64 unsigned-byte-63 fixnum positive-fixnum)))
       (32 (values '(unsigned-byte-32 unsigned-byte-31 positive-fixnum)
-                  '(signed-byte-32 fixnum unsigned-byte-31 positive-fixnum))))
+                  '(signed-byte-32 fixnum unsigned-byte-31 positive-fixnum)
+                  '(signed-byte-32 unsigned-byte-32 unsigned-byte-31 fixnum positive-fixnum))))
   (!def-primitive-type-alias unsigned-num `(:or ,@unsigned))
   (!def-primitive-type-alias signed-num `(:or ,@signed))
-  (!def-primitive-type-alias untagged-num
-    `(:or ,@(sort (copy-list (union unsigned signed)) #'string<))))
+  (!def-primitive-type-alias untagged-num `(:or ,@untagged)))
 
 ;;; other primitive immediate types
 (/show0 "primtype.lisp 68")
@@ -78,6 +78,7 @@
 
 ;;; primitive other-pointer number types
 (/show0 "primtype.lisp 81")
+(!def-primitive-type integer (descriptor-reg))
 (!def-primitive-type bignum (descriptor-reg))
 (!def-primitive-type ratio (descriptor-reg))
 (!def-primitive-type complex (descriptor-reg))
@@ -92,17 +93,73 @@
 (/show0 "about to !DEF-PRIMITIVE-TYPE COMPLEX-DOUBLE-FLOAT")
 (!def-primitive-type complex-double-float (complex-double-reg descriptor-reg)
   :type (complex double-float))
-#!+sb-simd-pack
+#+sb-simd-pack
 (progn
   (/show0 "about to !DEF-PRIMITIVE-TYPE SIMD-PACK")
   (!def-primitive-type simd-pack-single (single-sse-reg descriptor-reg)
     :type (simd-pack single-float))
   (!def-primitive-type simd-pack-double (double-sse-reg descriptor-reg)
     :type (simd-pack double-float))
-  (!def-primitive-type simd-pack-int (int-sse-reg descriptor-reg)
-   :type (simd-pack integer))
+  (!def-primitive-type simd-pack-ub8 (int-sse-reg descriptor-reg)
+    :type (simd-pack (unsigned-byte 8)))
+  (!def-primitive-type simd-pack-ub16 (int-sse-reg descriptor-reg)
+    :type (simd-pack (unsigned-byte 16)))
+  (!def-primitive-type simd-pack-ub32 (int-sse-reg descriptor-reg)
+    :type (simd-pack (unsigned-byte 32)))
+  (!def-primitive-type simd-pack-ub64 (int-sse-reg descriptor-reg)
+    :type (simd-pack (unsigned-byte 64)))
+  (!def-primitive-type simd-pack-sb8 (int-sse-reg descriptor-reg)
+    :type (simd-pack (signed-byte 8)))
+  (!def-primitive-type simd-pack-sb16 (int-sse-reg descriptor-reg)
+    :type (simd-pack (signed-byte 16)))
+  (!def-primitive-type simd-pack-sb32 (int-sse-reg descriptor-reg)
+    :type (simd-pack (signed-byte 32)))
+  (!def-primitive-type simd-pack-sb64 (int-sse-reg descriptor-reg)
+    :type (simd-pack (signed-byte 64)))
   (!def-primitive-type-alias simd-pack
-   '(:or simd-pack-single simd-pack-double simd-pack-int)))
+   '(:or simd-pack-single
+         simd-pack-double
+         simd-pack-ub8
+         simd-pack-ub16
+         simd-pack-ub32
+         simd-pack-ub64
+         simd-pack-sb8
+         simd-pack-sb16
+         simd-pack-sb32
+         simd-pack-sb64)))
+#+sb-simd-pack-256
+(progn
+  (!def-primitive-type simd-pack-256-single (single-avx2-reg descriptor-reg)
+    :type (simd-pack-256 single-float))
+  (!def-primitive-type simd-pack-256-double (double-avx2-reg descriptor-reg)
+    :type (simd-pack-256 double-float))
+  (!def-primitive-type simd-pack-256-ub8 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (unsigned-byte 8)))
+  (!def-primitive-type simd-pack-256-ub16 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (unsigned-byte 16)))
+  (!def-primitive-type simd-pack-256-ub32 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (unsigned-byte 32)))
+  (!def-primitive-type simd-pack-256-ub64 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (unsigned-byte 64)))
+  (!def-primitive-type simd-pack-256-sb8 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (signed-byte 8)))
+  (!def-primitive-type simd-pack-256-sb16 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (signed-byte 16)))
+  (!def-primitive-type simd-pack-256-sb32 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (signed-byte 32)))
+  (!def-primitive-type simd-pack-256-sb64 (int-avx2-reg descriptor-reg)
+    :type (simd-pack-256 (signed-byte 64)))
+  (!def-primitive-type-alias simd-pack-256
+   '(:or simd-pack-256-single
+         simd-pack-256-double
+         simd-pack-256-ub8
+         simd-pack-256-ub16
+         simd-pack-256-ub32
+         simd-pack-256-ub64
+         simd-pack-256-sb8
+         simd-pack-256-sb16
+         simd-pack-256-sb32
+         simd-pack-256-sb64)))
 
 ;;; primitive other-pointer array types
 (/show0 "primtype.lisp 96")
@@ -132,13 +189,18 @@
 ;;; Return the most restrictive primitive type that contains OBJECT.
 (/show0 "primtype.lisp 147")
 (defun primitive-type-of (object)
-  (let ((type (ctype-of object)))
-    (cond ((not (member-type-p type)) (primitive-type type))
-          ((and (eql 1 (member-type-size type))
-                (equal (member-type-members type) '(nil)))
-           (primitive-type-or-lose 'list))
-          (t
-           *backend-t-primitive-type*))))
+  (if (null object)
+      (load-time-value (primitive-type-or-lose 'list) t)
+      (let ((type (ctype-of object)))
+        (if (member-type-p type)
+            *backend-t-primitive-type*
+            (primitive-type type)))))
+
+#+sb-simd-pack
+(defun simd-pack-mask->tag (mask)
+  (aver (= (logcount mask) 1))
+  (the (integer 0 (#.(length +simd-pack-element-types+)))
+       (1- (integer-length mask))))
 
 ;;; Return the primitive type corresponding to a type descriptor
 ;;; structure. The second value is true when the primitive type is
@@ -152,7 +214,7 @@
 ;;; !DEF-VM-SUPPORT-ROUTINE and DEFUN-CACHED.
 (/show0 "primtype.lisp 188")
 (defun primitive-type (type)
-  (sb!kernel::maybe-reparse-specifier! type)
+  (sb-kernel::maybe-reparse-specifier! type)
   (primitive-type-aux type))
 (/show0 "primtype.lisp 191")
 (defun-cached (primitive-type-aux
@@ -224,7 +286,7 @@
                 (integer
                  (cond ((and hi lo)
                         (dolist (spec
-                                  `((positive-fixnum 0 ,sb!xc:most-positive-fixnum)
+                                  `((positive-fixnum 0 ,most-positive-fixnum)
                                     ,@(ecase n-machine-word-bits
                                         (32
                                          `((unsigned-byte-31
@@ -236,8 +298,8 @@
                                             0 ,(1- (ash 1 63)))
                                            (unsigned-byte-64
                                             0 ,(1- (ash 1 64))))))
-                                    (fixnum ,sb!xc:most-negative-fixnum
-                                            ,sb!xc:most-positive-fixnum)
+                                    (fixnum ,most-negative-fixnum
+                                            ,most-positive-fixnum)
                                     ,(ecase n-machine-word-bits
                                        (32
                                         `(signed-byte-32 ,(ash -1 31)
@@ -245,10 +307,10 @@
                                        (64
                                         `(signed-byte-64 ,(ash -1 63)
                                                          ,(1- (ash 1 63))))))
-                                 (if (or (< hi sb!xc:most-negative-fixnum)
-                                         (> lo sb!xc:most-positive-fixnum))
+                                 (if (or (< hi most-negative-fixnum)
+                                         (> lo most-positive-fixnum))
                                      (part-of bignum)
-                                     (any)))
+                                     (part-of integer)))
                           (let ((type (car spec))
                                 (min (cadr spec))
                                 (max (caddr spec)))
@@ -256,11 +318,11 @@
                               (return (values
                                        (primitive-type-or-lose type)
                                        (and (= lo min) (= hi max))))))))
-                       ((or (and hi (< hi sb!xc:most-negative-fixnum))
-                            (and lo (> lo sb!xc:most-positive-fixnum)))
+                       ((or (and hi (< hi most-negative-fixnum))
+                            (and lo (> lo most-positive-fixnum)))
                         (part-of bignum))
                        (t
-                        (any))))
+                        (part-of integer))))
                 (float
                  (let ((exact (and (null lo) (null hi))))
                    (case (numeric-type-format type)
@@ -320,37 +382,37 @@
                              (setq res new-ptype)
                              (return (any)))))))))))
         (intersection-type
-         (if (types-equal-or-intersect (specifier-type 'function) type)
-             ;; Things like (AND STANDARD-OBJECT FUNCTION) are callable as functions.
-             (part-of function)
-             (let ((types (intersection-type-types type))
-                   (res (any)))
-               ;; why NIL for the exact?  Well, we assume that the
-               ;; intersection type is in fact doing something for us:
-               ;; that is, that each of the types in the intersection is
-               ;; in fact cutting off some of the type lattice.  Since no
-               ;; intersection type is represented by a primitive type and
-               ;; primitive types are mutually exclusive, it follows that
-               ;; no intersection type can represent the entirety of the
-               ;; primitive type.  (And NIL is the conservative answer,
-               ;; anyway).  -- CSR, 2006-09-14
-               (dolist (type types (values res nil))
-                 (multiple-value-bind (ptype)
-                     (primitive-type type)
-                   (cond
-                     ;; if the result so far is (any), any improvement on
-                     ;; the specificity of the primitive type is valid.
-                     ((eq res (any))
-                      (setq res ptype))
-                     ;; if the primitive type returned is (any), the
-                     ;; result so far is valid.  Likewise, if the
-                     ;; primitive type is the same as the result so far,
-                     ;; everything is fine.
-                     ((or (eq ptype (any)) (eq ptype res)))
-                     ;; otherwise, we have something hairy and confusing,
-                     ;; such as (and condition funcallable-instance).
-                     ;; Punt.
-                     (t (return (any)))))))))
+         (let ((types (intersection-type-types type))
+               (res (any)))
+           ;; why NIL for the exact?  Well, we assume that the
+           ;; intersection type is in fact doing something for us:
+           ;; that is, that each of the types in the intersection is
+           ;; in fact cutting off some of the type lattice.  Since no
+           ;; intersection type is represented by a primitive type and
+           ;; primitive types are mutually exclusive, it follows that
+           ;; no intersection type can represent the entirety of the
+           ;; primitive type.  (And NIL is the conservative answer,
+           ;; anyway).  -- CSR, 2006-09-14
+           (dolist (type types (values res nil))
+             (when (csubtypep type (specifier-type 'function))
+               ;; Things like (AND STANDARD-OBJECT FUNCTION) are callable as functions.
+               (part-of function))
+             (multiple-value-bind (ptype)
+                 (primitive-type type)
+               (cond
+                 ;; if the result so far is (any), any improvement on
+                 ;; the specificity of the primitive type is valid.
+                 ((eq res (any))
+                  (setq res ptype))
+                 ;; if the primitive type returned is (any), the
+                 ;; result so far is valid.  Likewise, if the
+                 ;; primitive type is the same as the result so far,
+                 ;; everything is fine.
+                 ((or (eq ptype (any)) (eq ptype res)))
+                 ;; otherwise, we have something hairy and confusing,
+                 ;; such as (and condition funcallable-instance).
+                 ;; Punt.
+                 (t (return (any))))))))
         (member-type
          (let (res)
            (block nil
@@ -378,25 +440,30 @@
          (if (eq type (specifier-type 'character))
              (exactly character)
              (part-of character)))
-        #!+sb-simd-pack
+        #+sb-simd-pack
         (simd-pack-type
-         (let ((eltypes (simd-pack-type-element-type type)))
-           (cond ((member 'integer eltypes)
-                  (exactly simd-pack-int))
-                 ((member 'single-float eltypes)
-                  (exactly simd-pack-single))
-                 ((member 'double-float eltypes)
-                  (exactly simd-pack-double)))))
+         (let ((mask (simd-pack-type-tag-mask type)))
+           (if (= (logcount mask) 1)
+               (values (primitive-type-or-lose
+                        (svref +simd-pack-128-primtypes+ (simd-pack-mask->tag mask)))
+                       t)
+               (any))))
+        #+sb-simd-pack-256
+        (simd-pack-256-type
+         (let ((mask (simd-pack-256-type-tag-mask type)))
+           (if (= (logcount mask) 1)
+               (values (primitive-type-or-lose
+                        (svref +simd-pack-256-primtypes+ (simd-pack-mask->tag mask)))
+                       t)
+               (any))))
+        (cons-type
+         (part-of list))
         (built-in-classoid
          (case (classoid-name type)
-           #!+sb-simd-pack
-           ;; Can't tell what specific type; assume integers.
-           (simd-pack
-            (exactly simd-pack-int))
            ((complex function system-area-pointer weak-pointer)
             (values (primitive-type-or-lose (classoid-name type)) t))
-           (cons-type
-            (part-of list))
+           ((pathname logical-pathname)
+            (part-of instance))
            (t
             (any))))
         (fun-designator-type
