@@ -49,18 +49,22 @@
            (- list-pointer-lowtag)))
       0))
 
+(symbol-macrolet ((alien-linkage-table-space-end
+                   (+ alien-linkage-table-space-start alien-linkage-table-space-size)))
 ;;; the address of the linkage table entry for table index I.
-(defun linkage-table-entry-address (i)
-  (ecase linkage-table-growth-direction
-    (:up   (+ (* i linkage-table-entry-size) linkage-table-space-start))
-    (:down (- linkage-table-space-end (* (1+ i) linkage-table-entry-size)))))
+(defun alien-linkage-table-entry-address (i)
+  (ecase alien-linkage-table-growth-direction
+    (:up   (+ (* i alien-linkage-table-entry-size) alien-linkage-table-space-start))
+    (:down (- alien-linkage-table-space-end (* (1+ i) alien-linkage-table-entry-size)))))
 
-(defun linkage-table-index-from-address (addr)
-  (ecase linkage-table-growth-direction
+#-sb-xc-host
+(defun alien-linkage-table-index-from-address (addr)
+  (ecase alien-linkage-table-growth-direction
     (:up
-     (floor (- addr linkage-table-space-start) linkage-table-entry-size))
+     (floor (- addr alien-linkage-table-space-start) alien-linkage-table-entry-size))
     (:down
-     (1- (floor (- linkage-table-space-end addr) linkage-table-space-end)))))
+     (1- (floor (- alien-linkage-table-space-end addr) alien-linkage-table-space-end)))))
+)
 
 (defconstant-eqx +all-static-fdefns+
     #.(concatenate 'vector +c-callable-fdefns+ +static-fdefns+) #'equalp)
@@ -247,27 +251,44 @@
        (not (types-equal-or-intersect (tn-ref-type tn-ref)
                                       (specifier-type 'list)))))
 
-;;; Does the TN definitely hold an OTHER pointer
-(defun other-pointer-tn-ref-p (tn-ref)
+;;; Does the TN definitely hold an OTHER pointer?
+;;; If the operation next to be performed on TN is a widetag test,
+;;; then NIL is ok as the input. Indicate this by specifying PERMIT-NIL.
+;;; With rare exception it should always be permitted, though not on ppc64
+;;; where it would never be. The safe default is NIL.
+(defun other-pointer-tn-ref-p (tn-ref &optional permit-nil)
   (and (sc-is (tn-ref-tn tn-ref) descriptor-reg)
        (not (types-equal-or-intersect
              (tn-ref-type tn-ref)
-             (specifier-type '(or fixnum
-                               #+64-bit single-float
-                               function
-                               list
-                               instance
-                               character))))))
+             (if permit-nil
+                 (specifier-type '(or fixnum
+                                   #+64-bit single-float
+                                   function
+                                   cons
+                                   instance
+                                   character))
+                 (specifier-type '(or fixnum
+                                   #+64-bit single-float
+                                   function
+                                   list
+                                   instance
+                                   character)))))))
 
-(defun fixnum-or-other-pointer-tn-ref-p (tn-ref)
+(defun fixnum-or-other-pointer-tn-ref-p (tn-ref &optional permit-nil)
   (and (sc-is (tn-ref-tn tn-ref) descriptor-reg)
        (not (types-equal-or-intersect
              (tn-ref-type tn-ref)
-             (specifier-type '(or #+64-bit single-float
-                               function
-                               list
-                               instance
-                               character))))))
+             (specifier-type (if permit-nil
+                                 '(or #+64-bit single-float
+                                   function
+                                   cons
+                                   instance
+                                   character)
+                                 '(or #+64-bit single-float
+                                   function
+                                   list
+                                   instance
+                                   character)))))))
 
 (defun not-nil-tn-ref-p (tn-ref)
   (not (types-equal-or-intersect (tn-ref-type tn-ref)
@@ -392,3 +413,54 @@
 (defmacro id-bits-offset ()
   (let ((slot (get-dsd-index layout sb-kernel::id-word0)))
     (ash (+ sb-vm:instance-slots-offset slot) sb-vm:word-shift)))
+
+;;; I'd like the division-by-constant-integer optimization to work
+;;; during cross-compilation, but the algorithm to compute the magic
+;;; parameters is expressed in C, not Lisp. I need to translate it.
+#-sb-xc-host
+(defun sb-c:compute-udiv32-magic (divisor)
+  (with-alien ((mag (struct magu
+                            (m unsigned-int)
+                            (a int)
+                            (s int)))
+               (compute-udiv-magic32 (function void int (* (struct magu)))
+                                     :extern))
+    (alien-funcall compute-udiv-magic32 divisor (addr mag))
+    (values (slot mag 'm) (slot mag 'a) (slot mag 's))))
+
+;;; "Algorithm 2: Algorithm to select the number of fractional bits and the scaled
+;;; approximate reciprocal in the case of unsigned integers."
+;;; from https://r-libre.teluq.ca/1633/1/Faster_Remainder_of_the_Division_by_a_Constant.pdf
+;;; See also https://github.com/bmkessler/fastdiv for that coded in Go.
+;;; D = divisor
+;;; N = number of bits of precision in numerator
+;;; FRACTION-BITS is what you want, or :VARIABLE for the smallest
+;;;
+;;; Note that for 32 fraction bits, the divisor can *not* use all 32 bits of precision.
+;;; It can only have about 27 or 28 significant bits. This function will figure it out.
+(defun compute-fastrem-coefficient (d n fraction-bits)
+  (multiple-value-bind (smallest-f c)
+      (flet ((is-pow2 (n)
+               (declare (unsigned-byte n))
+               (let ((l (integer-length n)))
+                 (= n (ash 1 (1- l))))))
+        (if (is-pow2 d)
+            (values (1- (integer-length d)) 1)
+            (loop for L from 0
+                  do (let* ((F (+ N L))
+                            (2^F (expt 2 F)))
+                       (when (<= d (+ (mod 2^F d) (expt 2 L)))
+                         (let ((c (ceiling (expt 2 F) d)))
+                           (return (values F c))))))))
+    (cond ((eq fraction-bits :variable) ; return the smallest F
+           (values c smallest-f))
+          (t
+           ;; Otherwise hardwire F to 32 so the algorithm can use :DWORD
+           ;; register moves to perform the shifting and masking.
+           ;; But make sure the smallest-f is not more than 32, or else
+           ;; this can't work.
+           (when (> smallest-f fraction-bits)
+             (error "Need ~D fraction bits for divisor ~D and ~D bit dividend"
+                    smallest-f d n))
+           (values (ceiling (expt 2 fraction-bits) d) fraction-bits)))))
+
