@@ -56,14 +56,11 @@
            #+long-float simple-array-complex-long-float-p
            simple-rank-1-array-*-p
            system-area-pointer-p realp
-           ;; #-64-bit
+           signed-byte-8-p signed-byte-16-p
            unsigned-byte-32-p
-           ;; #-64-bit
            signed-byte-32-p
-           #+64-bit
-           unsigned-byte-64-p
-           #+64-bit
-           signed-byte-64-p
+           #+64-bit unsigned-byte-64-p
+           #+64-bit signed-byte-64-p
            weak-pointer-p code-component-p lra-p
            sb-int:unbound-marker-p
            pointerp
@@ -74,6 +71,8 @@
            non-null-symbol-p)
     (t) boolean (movable foldable flushable))
 
+(defknown car-eq-if-listp (t t) boolean (movable foldable flushable))
+
 (defknown #.(loop for (name) in *vector-without-complex-typecode-infos*
                   collect name)
     (t) boolean (movable foldable flushable))
@@ -95,17 +94,31 @@
 (defknown pointer-hash (t) fixnum (flushable))
 
 (defknown %sp-string-compare
-  (simple-string index (or null index) simple-string index (or null index))
+  (simple-string simple-string index (or null index) index (or null index))
   (values index fixnum)
-  (foldable flushable))
+  (foldable flushable no-verify-arg-count))
+(defknown %sp-string=
+  (simple-string simple-string index (or null index) index (or null index))
+  boolean
+  (foldable flushable no-verify-arg-count))
 
+(defknown (sb-impl::instance-sxhash sb-impl::%instance-sxhash)
+    (instance) hash-code (flushable))
+;;; SXHASH values on numbers and strings are predictable, therefore the next batch
+;;; of functions are flushable. Perhaps not entirely obviously, symbol hashes are
+;;; predictable because we hash by name. And while ENSURE-SYMBOL-HASH is not
+;;; - strictly speaking - a side-effectless operation, it has no effect as far as
+;;; user-written code is concerned. i.e. nothing portably expressed can discern
+;;; whether memoization of the hash occurred. (I feel like we would have a cleaner
+;;; implementation if we just eagerly compute the hash anyway. Practically all
+;;; symbols eventually get hashed)
 (defknown sb-impl::number-sxhash (number) hash-code (foldable flushable))
 (defknown %sxhash-string (string) hash-code (foldable flushable))
 (defknown %sxhash-simple-string (simple-string) hash-code (foldable flushable))
 
 (defknown (%sxhash-simple-substring) (simple-string index index) hash-code
   (foldable flushable))
-(defknown (compute-symbol-hash) (simple-string index) hash-code
+(defknown sb-impl::compute-symbol-hash (simple-string index) hash-code
   (foldable flushable))
 
 (defknown (symbol-hash ensure-symbol-hash) (symbol) hash-code
@@ -123,7 +136,15 @@
 (defknown %set-symbol-hash (symbol hash-code)
   t ())
 
-(defknown symbol-info-vector (symbol) (or null simple-vector))
+;;; SYMBOL-PACKAGE-ID for #+compact-symbol demands a vop which avoids loading
+;;; a raw bit value in a descriptor register (the SLOT vop returns a descriptor)
+(defknown symbol-package-id (symbol) (unsigned-byte 16))
+;;; TODO: I'd like to eliminate the (OR NULL) from this return type.
+;;; For that to happen, I probably need +nil-packed-infos+ to become
+;;; placed in static space because assembly routines may need it.
+;;; On the other hand, they may not, because there is no special case
+;;; code needed when reading from it, which is entire point.
+(defknown symbol-dbinfo (symbol) (or null packed-info))
 
 (defknown initialize-vector ((simple-array * (*)) &rest t)
   (simple-array * (*))
@@ -131,7 +152,7 @@
   :result-arg 0)
 
 (defknown (vector-fill* vector-fill/t) (t t t t) vector
-  ()
+  (no-verify-arg-count)
   :result-arg 0)
 
 ;;; Return the length of VECTOR.
@@ -158,51 +179,64 @@
 (defknown (function-header-word) (function) sb-vm:word (flushable))
 (defknown (instance-header-word) (instance) sb-vm:word (flushable))
 
-;;; This unconventional setter returns its first arg, not the newval.
 (defknown set-header-data
-    (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) t)
-;;; Perform a bitwise OR or AND-not of the existing header data with
-;;; the new bits and return no value.
-(defknown (set-header-bits unset-header-bits)
-  (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) (values)
-  ())
-(defknown (test-header-bit)
+    (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) (values))
+;;; Like SET-HEADER-DATA, but instead of writing the entire header,
+;;; LOGIOR of the specified value into the "data" portion of the word.
+;;; Returns the first argument, *not* the modified header data.
+(defknown logior-header-bits (t (unsigned-byte 16)) t
+    (#+x86-64 always-translatable))
+;;; ASSIGN-VECTOR-FLAGSS assign all and only the flags byte.
+;;; RESET- performs LOGANDC2 and returns no value.
+(defknown (assign-vector-flags reset-header-bits)
+  (t (unsigned-byte 16)) (values)
+  (#+x86-64 always-translatable))
+;;; test bits of "HeaderData" which start 8 bits over from the lsb
+(defknown (test-header-data-bit)
   (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) (boolean)
   (flushable))
 
 (defknown %array-dimension (array index) index
   (flushable))
-(defknown %set-array-dimension (array index index) index
+(defknown %set-array-dimension (array index index) (values)
   ())
 (defknown %array-rank (array) array-rank
   (flushable))
 
-#+(or x86 x86-64)
+#+(or x86 x86-64 arm64)
 (defknown (%array-rank= widetag=) (t t) boolean
   (flushable))
 
+(defknown simple-array-header-of-rank-p (t array-rank) boolean
+  (flushable))
 (defknown sb-kernel::check-array-shape (simple-array list)
   (simple-array)
-  (flushable)
+  (flushable no-verify-arg-count)
   :result-arg 0)
 
-(defknown %make-instance (index) instance
+(defknown (%make-instance %make-instance/mixed) (index) instance
   (flushable))
 (defknown %make-structure-instance (defstruct-description list &rest t) instance
   (flushable always-translatable))
 (defknown (%copy-instance %copy-instance-slots) (instance instance) instance
   () :result-arg 0)
-(defknown %instance-layout (instance) layout
-  (foldable flushable))
+(defknown %instance-layout (instance) sb-vm:layout (foldable flushable))
+(defknown %instance-wrapper (instance) wrapper (foldable flushable))
 ;;; %FUN-LAYOUT is to %INSTANCE-LAYOUT as FUN-POINTER-LOWTAG is to INSTANCE-POINTER-LOWTAG
 (defknown %fun-layout (#-compact-instance-header funcallable-instance
                        #+compact-instance-header function)
-  layout
+  sb-vm:layout (foldable flushable))
+(defknown %fun-wrapper (#-compact-instance-header funcallable-instance
+                        #+compact-instance-header function)
+  wrapper
   (foldable flushable))
-(defknown %set-instance-layout (instance layout) layout
-  ())
-(defknown %set-funcallable-instance-layout (funcallable-instance layout) layout
-  ())
+(defknown %set-instance-layout (instance sb-vm:layout) (values) ())
+;;; %SET-FUN-LAYOUT should only called on FUNCALLABLE-INSTANCE
+(defknown %set-fun-layout (funcallable-instance sb-vm:layout) (values) ())
+;;; Layout getter that accepts any object, and if it has INSTANCE- or FUN-
+;;; POINTER-LOWTAG returns the layout, otherwise some agreed-upon layout.
+(defknown %instanceoid-layout (t) sb-vm:layout (flushable))
+(defknown layout-eq ((or instance function) t (mod 16)) boolean (flushable))
 ;;; Caution: This is not exactly the same as instance_length() in C.
 ;;; The C one is the same as SB-VM::INSTANCE-LENGTH.
 (defknown %instance-length (instance) (unsigned-byte 14) (foldable flushable))
@@ -211,38 +245,31 @@
   (flushable always-translatable))
 (defknown (%instance-ref-eq) (instance index t) boolean
   (flushable always-translatable))
-(defknown %instance-set (instance index t) t
-  (always-translatable)
-  :derive-type #'result-type-last-arg)
-(defknown update-object-layout (t) layout)
+(defknown %instance-set (instance index t) (values) (always-translatable))
+(defknown update-object-layout (t) sb-vm:layout)
 
 #+(or arm64 ppc ppc64 riscv x86 x86-64)
 (defknown %raw-instance-cas/word (instance index sb-vm:word sb-vm:word)
   sb-vm:word ())
-#+riscv
+#+(or arm64 riscv x86 x86-64)
 (defknown %raw-instance-cas/signed-word (instance index sb-vm:signed-word sb-vm:signed-word)
   sb-vm:signed-word ())
+(defknown %raw-instance-xchg/word (instance index sb-vm:word) sb-vm:word ())
 
-#.`(progn
-     ,@(map 'list
-            (lambda (rsd)
-              (let* ((reader (sb-kernel::raw-slot-data-accessor-name rsd))
-                     (name (copy-seq (string reader)))
-                     (writer (intern (replace name "-SET/"
-                                              :start1 (search "-REF/" name))))
-                     (type (sb-kernel::raw-slot-data-raw-type rsd)))
-                `(progn
-                   (defknown ,reader (instance index) ,type
-                     (flushable always-translatable))
-                   (defknown ,writer (instance index ,type) ,type
-                     (always-translatable) :derive-type #'result-type-last-arg)
-                   ;; Interpreter stubs, harmless but unnecessary on host
-                   #-sb-xc-host
-                   (progn (defun ,reader (instance index)
-                            (,reader instance index))
-                          (defun ,writer (instance index new-value)
-                            (,writer instance index new-value))))))
-            sb-kernel::*raw-slot-data*))
+(macrolet ((define-raw-slot-defknowns ()
+             `(progn
+                ,@(map 'list
+                       (lambda (rsd)
+                         (let* ((reader (sb-kernel::raw-slot-data-reader-name rsd))
+                                (writer (sb-kernel::raw-slot-data-writer-name rsd))
+                                (type (sb-kernel::raw-slot-data-raw-type rsd)))
+                           `(progn
+                              (defknown ,reader (instance index) ,type
+                                  (flushable always-translatable))
+                              (defknown ,writer (instance index ,type) (values)
+                                  (always-translatable)))))
+                       sb-kernel::*raw-slot-data*))))
+  (define-raw-slot-defknowns))
 
 #+compare-and-swap-vops
 (defknown %raw-instance-atomic-incf/word (instance index sb-vm:word) sb-vm:word
@@ -254,15 +281,23 @@
 ;;; These two are mostly used for bit-bashing operations.
 (defknown %vector-raw-bits (t index) sb-vm:word
   (flushable))
-(defknown (%set-vector-raw-bits) (t index sb-vm:word) sb-vm:word
-  ())
+(defknown (%set-vector-raw-bits) (t index sb-vm:word) (values) ())
 
 
 ;;; Allocate an unboxed, non-fancy vector with type code TYPE, length LENGTH,
 ;;; and WORDS words long. Note: it is your responsibility to ensure that the
 ;;; relation between LENGTH and WORDS is correct.
-;;; The extra bit beyond N_WIDETAG_BITS is for the vector weakness flag.
-(defknown allocate-vector ((unsigned-byte 9) index
+;;; Note that in almost all situations the first argument (TYPE) is a constant.
+;;; There are only 3 places that it can be non-constant, and not at all
+;;; after self-build is complete. The three non-constant places are from:
+;;;  - ALLOCATE-VECTOR-WITH-WIDETAG in src/code/array
+;;;  - ALLOCATE-VECTOR (which is basically a stub) in src/code/array
+;;;  - FOP-SPEC-VECTOR which calls allocate-vector
+;;; Apart from those, user code that does not have a compile-time-determined
+;;; vector type will usuallly end up calling allocate-vector-with-widetag
+;;; via %MAKE-ARRAY.
+(defknown allocate-vector (#+ubsan boolean
+                           word index
                            ;; The number of words is later converted
                            ;; to bytes, make sure it fits.
                            (and index
@@ -317,14 +352,30 @@
       (flushable movable foldable))
   (defknown %make-simd-pack-ub32 ((unsigned-byte 32) (unsigned-byte 32)
                                   (unsigned-byte 32) (unsigned-byte 32))
-      (simd-pack integer)
-      (flushable movable foldable))
+      (simd-pack (unsigned-byte 32))
+      (flushable movable #-sb-xc-host foldable))
   (defknown %make-simd-pack-ub64 ((unsigned-byte 64) (unsigned-byte 64))
-      (simd-pack integer)
-      (flushable movable foldable))
+      (simd-pack (unsigned-byte 64))
+      (flushable movable #-sb-xc-host foldable))
   (defknown (%simd-pack-low %simd-pack-high) (simd-pack)
       (unsigned-byte 64)
       (flushable movable foldable))
+  (defknown %simd-pack-ub8s (simd-pack)
+      (values (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-ub16s (simd-pack)
+      (values (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16))
+      (flushable movable foldable))
   (defknown %simd-pack-ub32s (simd-pack)
       (values (unsigned-byte 32) (unsigned-byte 32)
               (unsigned-byte 32) (unsigned-byte 32))
@@ -332,6 +383,29 @@
   (defknown %simd-pack-ub64s (simd-pack)
       (values (unsigned-byte 64) (unsigned-byte 64))
       (flushable movable foldable))
+  (defknown %simd-pack-sb8s (simd-pack)
+      (values (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-sb16s (simd-pack)
+      (values (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16))
+      (flushable movable foldable))
+  (defknown %simd-pack-sb32s (simd-pack)
+      (values (signed-byte 32) (signed-byte 32)
+              (signed-byte 32) (signed-byte 32))
+      (flushable movable foldable))
+  (defknown %simd-pack-sb64s (simd-pack)
+      (values (signed-byte 64) (signed-byte 64))
+      (flushable movable foldable))
   (defknown %simd-pack-singles (simd-pack)
       (values single-float single-float single-float single-float)
       (flushable movable foldable))
@@ -356,14 +430,30 @@
       (flushable movable foldable))
   (defknown %make-simd-pack-256-ub32 ((unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32)
                                       (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32))
-      (simd-pack-256 integer)
+      (simd-pack-256 (unsigned-byte 32))
       (flushable movable foldable))
   (defknown %make-simd-pack-256-ub64 ((unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64))
-      (simd-pack-256 integer)
+      (simd-pack-256 (unsigned-byte 64))
       (flushable movable foldable))
   (defknown (%simd-pack-256-0 %simd-pack-256-1 %simd-pack-256-2 %simd-pack-256-3) (simd-pack-256)
       (unsigned-byte 64)
       (flushable movable foldable))
+  (defknown %simd-pack-256-ub8s (simd-pack-256)
+      (values (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-ub16s (simd-pack-256)
+      (values (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16))
+      (flushable movable foldable))
   (defknown %simd-pack-256-ub32s (simd-pack-256)
       (values (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32)
               (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32))
@@ -371,6 +461,29 @@
   (defknown %simd-pack-256-ub64s (simd-pack-256)
       (values (unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64))
       (flushable movable foldable))
+  (defknown %simd-pack-256-sb8s (simd-pack-256)
+      (values (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-sb16s (simd-pack-256)
+      (values (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-sb32s (simd-pack-256)
+      (values (signed-byte 32) (signed-byte 32) (signed-byte 32) (signed-byte 32)
+              (signed-byte 32) (signed-byte 32) (signed-byte 32) (signed-byte 32))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-sb64s (simd-pack-256)
+      (values (signed-byte 64) (signed-byte 64) (signed-byte 64) (signed-byte 64))
+      (flushable movable foldable))
   (defknown %simd-pack-256-singles (simd-pack-256)
       (values single-float single-float single-float single-float
               single-float single-float single-float single-float)
@@ -398,8 +511,10 @@
   system-area-pointer (flushable))
 (defknown (current-sp current-fp) () system-area-pointer (flushable))
 (defknown current-fp-fixnum () fixnum (flushable))
+;; STACK-REF is pretty nearly just SAP-REF-LISPOBJ except that
+;; it takes a word index, not a byte displacement from the SAP.
 (defknown stack-ref (system-area-pointer index) t (flushable))
-(defknown %set-stack-ref (system-area-pointer index t) t ())
+(defknown %set-stack-ref (system-area-pointer index t) (values) ())
 (defknown lra-code-header (t) t (movable flushable))
 ;; FUN-CODE-HEADER returns NIL for assembly routines that have a simple-fun header
 ;; with 0 as the data value. We should probably ensure that assembly routines
@@ -429,62 +544,102 @@
 ;;;; bignum operations
 
 (defknown %allocate-bignum (bignum-length) bignum
-  (flushable))
+  (flushable #-bignum-assertions always-translatable))
 
-(defknown %bignum-length (bignum) bignum-length
-  (foldable flushable))
+(defknown %bignum-length (bignum) (and (integer 1) bignum-length)
+  (foldable flushable always-translatable))
 
-(defknown %bignum-set-length (bignum bignum-length) bignum
-  ())
+;;; Change the length of bignum to be newlen. Newlen must be the same or
+;;; smaller than the old length, and any elements beyond newlen must be zeroed.
+(defknown %bignum-set-length (bignum bignum-length) (values)
+  (always-translatable))
 
 (defknown %bignum-ref (bignum bignum-index) bignum-element-type
-  (flushable))
+  (flushable #-bignum-assertions always-translatable))
+
 #+(or x86 x86-64)
 (defknown %bignum-ref-with-offset (bignum fixnum (signed-byte 24))
   bignum-element-type (flushable always-translatable))
 
-(defknown %bignum-set (bignum bignum-index bignum-element-type)
-  bignum-element-type
-  ())
+(defknown (%bignum-set #+bignum-assertions %%bignum-set) (bignum bignum-index bignum-element-type)
+  (values)
+  (#-bignum-assertions always-translatable))
 
+(defknown %half-bignum-ref (bignum sb-kernel::half-bignum-index) sb-kernel::half-bignum-element-type
+    (flushable #-bignum-assertions always-translatable))
+
+(defknown (setf %half-bignum-ref) (sb-kernel::half-bignum-element-type bignum sb-kernel::half-bignum-index)
+    (values)
+    (#-bignum-assertions always-translatable))
+
+;;; Return T if digit is positive, or NIL if negative.
 (defknown %digit-0-or-plusp (bignum-element-type) boolean
-  (foldable flushable movable))
+    (foldable flushable movable always-translatable))
 
+;;; %ADD-WITH-CARRY returns a bignum digit and a carry resulting from adding
+;;; together a, b, and an incoming carry.
+;;; %SUBTRACT-WITH-BOROW returns a bignum digit and a borrow resulting from
+;;; subtracting b from a, and subtracting a possible incoming borrow.
 (defknown (%add-with-carry %subtract-with-borrow)
-          (bignum-element-type bignum-element-type (mod 2))
-  (values bignum-element-type (mod 2))
-  (foldable flushable movable))
+    (bignum-element-type bignum-element-type (mod 2))
+    (values bignum-element-type (mod 2))
+    (foldable flushable movable always-translatable))
 
+;;; This multiplies x-digit and y-digit, producing high and low digits
+;;; manifesting the result. Then it adds the low digit, res-digit, and
+;;; carry-in-digit. Any carries (note, you still have to add two digits
+;;; at a time possibly producing two carries) from adding these three
+;;; digits get added to the high digit from the multiply, producing the
+;;; next carry digit.  Res-digit is optional since two uses of this
+;;; primitive multiplies a single digit bignum by a multiple digit
+;;; bignum, and in this situation there is no need for a result buffer
+;;; accumulating partial results which is where the res-digit comes
+;;; from.
 (defknown %multiply-and-add
-          (bignum-element-type bignum-element-type bignum-element-type
-                               &optional bignum-element-type)
-  (values bignum-element-type bignum-element-type)
-  (foldable flushable movable))
+    (bignum-element-type bignum-element-type bignum-element-type
+                         &optional bignum-element-type)
+    (values bignum-element-type bignum-element-type)
+    (foldable flushable movable always-translatable))
 
+;;; Multiply two digit-size numbers, returning a 2*digit-size result
+;;; split into two digit-size quantities.
 (defknown %multiply (bignum-element-type bignum-element-type)
-  (values bignum-element-type bignum-element-type)
-  (foldable flushable movable))
+    (values bignum-element-type bignum-element-type)
+    (foldable flushable movable always-translatable))
 
 (defknown %lognot (bignum-element-type) bignum-element-type
-  (foldable flushable movable))
+    (foldable flushable movable always-translatable))
 
 (defknown (%logand %logior %logxor) (bignum-element-type bignum-element-type)
   bignum-element-type
   (foldable flushable movable))
 
 (defknown %fixnum-to-digit (fixnum) bignum-element-type
-  (foldable flushable movable))
+  (foldable flushable movable #-(or arm arm64) always-translatable))
 
+;;; This takes three digits and returns the FLOOR'ed result of
+;;; dividing the first two as a 2*digit-size integer by the third.
 (defknown %bigfloor (bignum-element-type bignum-element-type bignum-element-type)
   (values bignum-element-type bignum-element-type)
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
+(defknown %half-bigfloor (half-bignum-element-type half-bignum-element-type half-bignum-element-type)
+  (values half-bignum-element-type half-bignum-element-type)
+  (foldable flushable movable always-translatable))
+
+;;; Convert the input, a BIGNUM-ELEMENT-TYPE, to a signed word.
+;;; FIXME: considering that both the input and output have N-WORD-BITS significant bits,
+;;; this is really a bad name for the operation.
 (defknown %fixnum-digit-with-correct-sign (bignum-element-type) sb-vm:signed-word
-  (foldable flushable movable))
+    (foldable flushable movable always-translatable))
 
+;;; %ASHR- take a digit-size quantity and shift it to the left,
+;;; returning a digit-size quantity.
+;;; %ASHR- Do an arithmetic shift right of data even though bignum-element-type is
+;;; unsigned.
 (defknown (%ashl %ashr %digit-logical-shift-right)
           (bignum-element-type (mod #.sb-vm:n-word-bits)) bignum-element-type
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
 ;;;; bit-bashing routines
 
@@ -528,13 +683,13 @@
 ;;; Extract the INDEXth element from the header of CODE-OBJ. Can be
 ;;; set with SETF.
 (defknown code-header-ref (code-component index) t (flushable))
-(defknown code-header-set (code-component index t) t ())
+(defknown code-header-set (code-component index t) (values) ())
 ;;; Extract a 4-byte element relative to the end of CODE-OBJ.
 ;;; The index should be strictly negative and a multiple of 4.
 (defknown code-trailer-ref (code-component fixnum) (unsigned-byte 32)
   (flushable #-(or sparc ppc64) always-translatable))
 
-(defknown fun-subtype (function) (member . #.sb-vm::+function-widetags+)
+(defknown %fun-pointer-widetag (function) (member . #.sb-vm::+function-widetags+)
   (flushable))
 
 (defknown make-fdefn (t) fdefn (flushable movable))
@@ -542,17 +697,18 @@
 (defknown fdefn-name (fdefn) t (foldable flushable))
 (defknown fdefn-fun (fdefn) (or function null) (flushable))
 (defknown (setf fdefn-fun) (function fdefn) t ())
-(defknown fdefn-makunbound (fdefn) t ())
+(defknown fdefn-makunbound (fdefn) (values) ())
 ;;; FDEFN -> FUNCTION, trapping if not FBOUNDP
 (defknown safe-fdefn-fun (fdefn) function ())
 
 (defknown %simple-fun-type (function) t (flushable))
 
-#+(or x86 x86-64) (defknown sb-vm::%closure-callee (function) fixnum (flushable))
+#+(or x86 x86-64 arm64) (defknown sb-vm::%closure-callee (function) fixnum (flushable))
 (defknown %closure-fun (function) function (flushable))
 
 (defknown %closure-index-ref (function index) t
   (flushable))
+(defknown %closure-index-set (function index t) (values) ())
 
 ;; T argument is for the 'fun' slot.
 (defknown sb-vm::%alloc-closure (index t) function (flushable))
@@ -563,7 +719,8 @@
   ())
 
 (defknown %funcallable-instance-info (function index) t (flushable))
-(defknown %set-funcallable-instance-info (function index t) t ())
+(defknown (setf %funcallable-instance-info) (t function index) t ())
+(defknown %set-funcallable-instance-info (function index t) (values) ())
 
 #+sb-fasteval
 (defknown sb-interpreter:fun-proto-fn (interpreted-function)
@@ -578,8 +735,15 @@
 
 ;;; formerly in 'float-tran'
 
-(defknown %single-float (real) single-float (movable foldable))
-(defknown %double-float (real) double-float (movable foldable))
+(defknown %single-float (real) single-float
+  (movable foldable no-verify-arg-count))
+(defknown %double-float (real) double-float
+  (movable foldable no-verify-arg-count))
+
+(defknown bignum-to-float (bignum symbol) float
+  (movable foldable no-verify-arg-count))
+(defknown sb-kernel::float-ratio (ratio symbol) float
+  (movable foldable no-verify-arg-count))
 
 (defknown make-single-float ((signed-byte 32)) single-float
   (movable flushable))
@@ -660,4 +824,21 @@
   (movable foldable flushable))
 
 (defknown (%unary-truncate %unary-round) (real) integer
-  (movable foldable flushable))
+  (movable foldable flushable no-verify-arg-count))
+
+#+(or arm64 x86-64)
+(defknown sb-lockless::get-next (sb-lockless::list-node) (values sb-lockless::list-node t))
+
+(defknown sb-vm::fastrem-32 ((unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32))
+  (unsigned-byte 32)
+  (flushable))
+(defknown sb-vm::fastrem-64 ((unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64))
+  (unsigned-byte 64)
+  (flushable))
+
+;;; +-modfx is useful for computing a hash that is commutative in its inputs.
+;;; These architectures lack a complete set of modular operations; they have operations
+;;; which are modular in N-WORD-BITS, but not that accept fixnums.
+#+(or arm mips ppc sparc)
+(defknown sb-vm::+-modfx (integer integer) fixnum
+          (movable foldable flushable always-translatable))
