@@ -468,6 +468,35 @@ body, references to a NAME will effectively be replaced with the EXPANSION."
                                   (subseq args required min))
                                ,@(subseq args 0 required)
                                ,@(subseq args min)))))
+
+(defmacro inline-%primitive (template &rest args)
+  (let* ((required (length (template-arg-types template)))
+         (info (template-info-arg-count template))
+         (min (+ required info))
+         (nargs (length args)))
+    (if (template-more-args-type template)
+        (when (< nargs min)
+          (bug "Primitive was called with ~R argument~:P, ~
+                but wants at least ~R."
+               nargs
+               min))
+        (unless (= nargs min)
+          (bug "Primitive was called with ~R argument~:P, ~
+                but wants exactly ~R."
+               nargs
+               min)))
+
+    (when (template-conditional-p template)
+      (bug "%PRIMITIVE was used with a conditional template."))
+
+    (when (template-more-results-type template)
+      (bug "%PRIMITIVE was used with an unknown values template."))
+
+    `(%%primitive ',template
+                  ',(eval-info-args
+                     (subseq args required min))
+                  ,@(subseq args 0 required)
+                  ,@(subseq args min))))
 
 ;;;; QUOTE
 
@@ -713,6 +742,12 @@ be a lambda expression."
 (def-ir1-translator %funcall-lvar ((function &rest args) start next result)
   (ir1-convert-combination-args function start next result args))
 
+(def-ir1-translator %funcall-no-nargs ((function &rest args) start next result)
+  (let ((ctran (make-ctran))
+        (fun-lvar (make-lvar)))
+    (ir1-convert start ctran fun-lvar `(the function ,function))
+    (ir1-convert-combination-args fun-lvar ctran next result args :pass-nargs nil)))
+
 ;;; This source transform exists to reduce the amount of work for the
 ;;; compiler. If the called function is a FUNCTION form, then convert
 ;;; directly to %FUNCALL, instead of waiting around for type
@@ -815,7 +850,8 @@ have been evaluated."
                                    :debug-name (debug-name 'let bindings))))
                          (reference-leaf start ctran fun-lvar fun))
                        (values next result))))
-           (ir1-convert-combination-args fun-lvar ctran next result values)))))
+           (ir1-convert-combination-args fun-lvar ctran next result values
+                                         :arg-source-forms bindings)))))
 
 (def-ir1-translator let* ((bindings &body body)
                           start next result)
@@ -826,7 +862,8 @@ form to reference any of the previous VARS."
   (multiple-value-bind (vars values forms decls) (parse-letish bindings body 'let*)
     (processing-decls (decls vars nil next result post-binding-lexenv)
       (ir1-convert-aux-bindings
-       start next result forms vars values post-binding-lexenv))))
+       start next result forms vars values post-binding-lexenv
+       :value-source-forms bindings))))
 
 ;;; logic shared between IR1 translators for LOCALLY, MACROLET,
 ;;; and SYMBOL-MACROLET
@@ -1197,6 +1234,7 @@ care."
       (let* ((name (first things))
              (value-form (second things))
              (leaf (or (lexenv-find name vars) (find-free-var name))))
+        (maybe-note-undefined-variable-reference leaf name)
         (etypecase leaf
           (leaf
            (when (constant-p leaf)
