@@ -260,7 +260,7 @@
   ;; This is only meaningful in :ARGUMENT and :TEMPORARY operands.
   (target nil :type (or symbol null) :read-only t)
   ;; TEMP is a temporary that holds the TN-REF for this operand.
-  (temp (make-operand-parse-temp) :type symbol :read-only t)
+  (temp (make-operand-parse-temp) :type symbol)
   ;; the time that this operand is first live and the time at which it
   ;; becomes dead again. These are TIME-SPECs, as returned by
   ;; PARSE-TIME-SPEC.
@@ -276,7 +276,8 @@
   ;; this operand is allowed into. If NIL, there is no restriction.
   (scs nil :type (or symbol list) :read-only t)
   ;; If non-null, we are a temp wired to this offset in SC.
-  (offset nil :type (or unsigned-byte null) :read-only t))
+  (offset nil :type (or unsigned-byte null) :read-only t)
+  (unused-if nil))
 (declaim (freeze-type operand-parse))
 
 (defun operand-parse-sc (parse) ; Enforce a single symbol
@@ -341,8 +342,8 @@
   (save-p nil :type (member t nil :compute-only :force-to-stack))
   ;; info about how to emit MOVE-ARG VOPs for the &MORE operand in
   ;; call/return VOPs
-  (move-args nil :type (member nil :local-call :full-call :known-return))
-  (args-var '.args. :type symbol)
+  (move-args nil :type (member nil :local-call :full-call :known-return :fixed))
+  (args-var nil :type list)
   (results-var '.results. :type symbol)
   (before-load :unspecified :type (or (member :unspecified) list)))
 (declaim (freeze-type vop-parse))
@@ -726,7 +727,6 @@
 (defun make-generator-function (parse)
   (declare (type vop-parse parse))
   (let ((n-vop (vop-parse-vop-var parse))
-        (n-args (vop-parse-args-var parse))
         (n-results (vop-parse-results-var parse))
         (operands (vop-parse-operands parse))
         (n-info (gensym)) (n-variant (gensym))
@@ -753,11 +753,10 @@
           ((:more-argument :more-result))))
 
       `(named-lambda (vop ,(vop-parse-name parse)) (,n-vop)
-         (let* ((,n-args (vop-args ,n-vop))
-                (,n-results (vop-results ,n-vop))
+         (let* ((,n-results (vop-results ,n-vop))
                 ,@(access-operands (vop-parse-args parse)
                                    (vop-parse-more-args parse)
-                                   n-args)
+                                   `(vop-args ,n-vop))
                 ,@(access-operands (vop-parse-results parse)
                                    (vop-parse-more-results parse)
                                    n-results)
@@ -780,13 +779,67 @@
            (declare (ignore ,@(vop-parse-ignores parse)
                             ,@(and (neq (vop-parse-before-load parse) :unspecified)
                                    `(,dummy)))
-                    (ignorable ,n-args ,n-results))
+                    (ignorable ,n-results))
            ,@(loads)
            ;; RETURN-FROM can exit the ASSEMBLE while continuing on with saves.
            (block ,(vop-parse-name parse)
              (assemble ()
                ,@(vop-parse-body parse)))
            ,@(saves))))))
+
+(defun make-after-sc-function (parse)
+  (let ((unused-temps
+          (remove-if-not #'operand-parse-unused-if
+                         (vop-parse-temps parse)))
+        (conditional (typep (vop-parse-conditional-p parse) '(cons (eql :after-sc-selection)))))
+    (when (or unused-temps
+              conditional)
+      (let* ((n-vop (vop-parse-vop-var parse))
+             (n-results (vop-parse-results-var parse))
+             (n-info (gensym))
+             (n-variant (gensym))
+             (bindings
+               `((,n-results (vop-results ,n-vop))
+                 ,@(access-operands (vop-parse-args parse)
+                                    (vop-parse-more-args parse)
+                                    `(vop-args ,n-vop))
+                 ,@(access-operands (vop-parse-results parse)
+                                    (vop-parse-more-results parse)
+                                    n-results)
+                 ,@(and unused-temps
+                        (access-operands (vop-parse-temps parse) nil
+                                         `(vop-temps ,n-vop)))
+                 ,@(when (vop-parse-info-args parse)
+                     `((,n-info (vop-codegen-info ,n-vop))
+                       ,@(mapcar (lambda (x) `(,x (pop ,n-info)))
+                                 (vop-parse-info-args parse))))
+                 ,@(when (vop-parse-variant-vars parse)
+                     `((,n-variant (vop-info-variant (vop-info ,n-vop)))
+                       ,@(mapcar (lambda (x) `(,x (pop ,n-variant)))
+                                 (vop-parse-variant-vars parse))))
+                 ,@(loop for op in (vop-parse-operands parse)
+                         when
+                         (ecase (operand-parse-kind op)
+                           ((:argument :result)
+                            `(,(operand-parse-name op)
+                              (tn-ref-tn ,(operand-parse-temp op))))
+                           (:temporary
+                            (and (operand-parse-unused-if op)
+                                 `(,(operand-parse-name op)
+                                   (tn-ref-tn ,(operand-parse-temp op)))))
+                           ((:more-argument :more-result)))
+                         collect it))))
+        `(lambda (,n-vop)
+           (let* ,bindings
+             (declare (ignorable ,@(mapcar #'car bindings)))
+             ,@(when conditional
+                 `((setf (car (last (vop-codegen-info ,n-vop)))
+                         (progn ,@(cdr (vop-parse-conditional-p parse))))
+                   (setf (vop-codegen-info ,n-vop)
+                         (butlast (vop-codegen-info ,n-vop)))))
+             ,@(loop for op in unused-temps
+                     collect `(when ,(operand-parse-unused-if op)
+                                (setf (tn-kind ,(operand-parse-name op)) :unused)))))))))
 
 (defvar *parse-vop-operand-count*)
 (defun make-operand-parse-temp ()
@@ -815,31 +868,35 @@
         (when more
           (error "The MORE operand isn't the last operand: ~S" specs))
         (incf *parse-vop-operand-count*)
+        (incf num)
         (let* ((name (first spec))
                (old (if (vop-parse-inherits parse)
                         (find-operand name
                                       (vop-parse-or-lose
                                        (vop-parse-inherits parse))
-                                      (list kind)
+                                      (list* kind
+                                             (if (eq kind :argument)
+                                                 '(:more-argument)
+                                                 '(:more-result)))
                                       nil)
                         nil))
                (res
-                (nconc (list :kind kind)
-                       (if old
-                        (list
-                         :target (operand-parse-target old)
-                         :born (operand-parse-born old)
-                         :dies (operand-parse-dies old)
-                         :scs (operand-parse-scs old)
-                         :load-tn (operand-parse-load-tn old)
-                         :load (operand-parse-load old))
-                        (ecase kind
-                          (:argument
-                           (list :born (parse-time-spec :load)
-                                 :dies (parse-time-spec `(:argument ,(incf num)))))
-                          (:result
-                           (list :born (parse-time-spec `(:result ,(incf num)))
-                                 :dies (parse-time-spec :save))))))))
+                 (nconc (list :kind kind)
+                        (if old
+                            (list
+                             :target (operand-parse-target old)
+                             :born (operand-parse-born old)
+                             :dies (operand-parse-dies old)
+                             :scs (operand-parse-scs old)
+                             :load-tn (operand-parse-load-tn old)
+                             :load (operand-parse-load old))
+                            (ecase kind
+                              (:argument
+                               (list :born (parse-time-spec :load)
+                                     :dies (parse-time-spec `(:argument ,num))))
+                              (:result
+                               (list :born (parse-time-spec `(:result ,num))
+                                     :dies (parse-time-spec :save))))))))
           (do ((tail (rest spec) (cddr tail)))
               ((null tail))
             (let ((key (first tail))
@@ -900,6 +957,11 @@
     (dolist (name (cddr spec))
       (unless (symbolp name)
         (error "bad temporary name: ~S" name))
+      ;; It's almost always a mistake to have overlaps in the operand names.
+      ;; But I guess that some users think it's fine?
+      #+sb-xc-host
+      (when (member name (vop-parse-temps parse) :key #'operand-parse-name)
+        (warn "temp ~s already exists in ~s" name (vop-parse-name parse)))
       (incf *parse-vop-operand-count*)
       (let ((res (list :born (parse-time-spec :load)
                        :dies (parse-time-spec :save))))
@@ -924,6 +986,7 @@
                (unless (= (length scs) 1)
                  (error "must specify exactly one SC for a temporary"))
                (setf value (first scs))))
+            (:unused-if)
             (t
              (error "unknown temporary option: ~S" opt)))
           (setf (getf res key) value)))
@@ -1008,13 +1071,13 @@
         (:vop-var
          (setf (vop-parse-vop-var parse) (vop-spec-arg spec 'symbol)))
         (:args-var
-         (setf (vop-parse-args-var parse) (vop-spec-arg spec 'symbol)))
+         (setf (vop-parse-args-var parse) (cdr spec)))
         (:results-var
          (setf (vop-parse-results-var parse) (vop-spec-arg spec 'symbol)))
         (:move-args
          (setf (vop-parse-move-args parse)
                (vop-spec-arg spec '(member nil :local-call :full-call
-                                    :known-return))))
+                                    :known-return :fixed))))
         (:node-var
          (setf (vop-parse-node-var parse) (vop-spec-arg spec 'symbol)))
         (:note
@@ -1047,6 +1110,11 @@
                        (rest spec))))
         (t
          (error "unknown option specifier: ~S" (first spec)))))
+    (loop with refs = (vop-parse-args-var parse)
+          for arg in (vop-parse-args parse)
+          for ref = (pop refs)
+          when ref
+          do (setf (operand-parse-temp arg) ref))
     (values)))
 
 ;;;; making costs and restrictions
@@ -1102,11 +1170,11 @@
     (values costs load-scs)))
 
 (defconstant-eqx +no-costs+
-    #.(make-array sb-vm:sc-number-limit :initial-element 0)
+  (make-array sb-vm:sc-number-limit :initial-element 0)
   #'equalp)
 
 (defconstant-eqx +no-loads+
-    #.(make-array sb-vm:sc-number-limit :initial-element t)
+  (make-array sb-vm:sc-number-limit :initial-element t)
   #'equalp)
 
 ;;; Pick off the case of operands with no restrictions.
@@ -1131,25 +1199,26 @@
       (compute-costs-and-restrictions-list (vop-parse-args parse) t)
     (multiple-value-bind (result-costs result-scs)
         (compute-costs-and-restrictions-list (vop-parse-results parse) nil)
-      `(
-        :cost ,(vop-parse-cost parse)
+      (multiple-value-bind (more-arg-costs more-arg-scs)
+          (and (vop-parse-more-args parse)
+               (compute-loading-costs-if-any (vop-parse-more-args parse) t))
+          `(:cost ,(vop-parse-cost parse)
 
-        :arg-costs ',arg-costs
-        :arg-load-scs ',arg-scs
-        :result-costs ',result-costs
-        :result-load-scs ',result-scs
+            :arg-costs ',arg-costs
+            :arg-load-scs ',arg-scs
+            :result-costs ',result-costs
+            :result-load-scs ',result-scs
 
-        :more-arg-costs
-        ',(if (vop-parse-more-args parse)
-              (compute-loading-costs-if-any (vop-parse-more-args parse) t)
-              nil)
+            :more-arg-costs ',more-arg-costs
+            :more-arg-load-scs ',(unless (eq more-arg-costs +no-loads+)
+                                   (substitute-if nil #'listp more-arg-scs))
 
-        :more-result-costs
-        ',(if (vop-parse-more-results parse)
-              (compute-loading-costs-if-any (vop-parse-more-results parse) nil)
-              nil)
-        :optional-results ',(loop for name in (vop-parse-optional-results parse)
-                                collect (position name (vop-parse-results parse) :key #'operand-parse-name))))))
+            :more-result-costs
+            ',(if (vop-parse-more-results parse)
+                  (compute-loading-costs-if-any (vop-parse-more-results parse) nil)
+                  nil)
+            :optional-results ',(loop for name in (vop-parse-optional-results parse)
+                                      collect (position name (vop-parse-results parse) :key #'operand-parse-name)))))))
 
 ;;;; operand checking and stuff
 
@@ -1232,7 +1301,9 @@
                       (when (sc-allowed-by-primitive-type
                              (sc-or-lose sc)
                              (primitive-type-or-lose ptype))
-                        (return t))))
+                        (return t)))
+                    #+arm64
+                    (eq sc 'sb-vm::zero))
           (warn "~:[Result~;Argument~] ~A to VOP ~S~@
                  has SC restriction ~S which is ~
                  not allowed by the operand type:~%  ~S"
@@ -1349,6 +1420,8 @@
       :more-args-type ,(when more-args (make-operand-type more-arg))
       :result-types ,(cond ((eq conditional t)
                             :conditional)
+                           ((eql (car conditional) :after-sc-selection)
+                            ''(:conditional :after-sc-selection))
                            (conditional
                             `'(:conditional . ,conditional))
                            (t
@@ -1411,7 +1484,10 @@
                (equal (vop-parse-body parse) (vop-parse-body iparse)))
           (unless (eq (vop-parse-body parse) :unspecified)
             (make-generator-function parse)))
-      :variant (list ,@variant))))
+      :variant (list ,@variant)
+      :after-sc-selection
+      ;; TODO: inherit it?
+      ,(make-after-sc-function parse))))
 
 ;;; Define the symbol NAME to be a Virtual OPeration in the compiler.
 ;;; If specified, INHERITS is the name of a VOP that we default
@@ -1573,6 +1649,9 @@
 ;;;     Indicates if and how the more args should be moved into a
 ;;;     different frame.
 (defmacro define-vop ((&optional name inherits) &body specs)
+  (%define-vop name inherits specs t))
+
+(defun %define-vop (name inherits specs set)
   (declare (type symbol name))
   ;; Parse the syntax into a VOP-PARSE structure, and then expand into
   ;; code that creates the appropriate VOP-INFO structure at load time.
@@ -1588,8 +1667,9 @@
       (let ((clause (assoc :translate specs)))
         (when (singleton-p (cdr clause))
           (setf name (cadr clause)))))
-    (aver (typep name '(and symbol (not null))))
-    (setf (vop-parse-name parse) name)
+    (when set
+      (aver (typep name '(and symbol (not null))))
+      (setf (vop-parse-name parse) name))
     (setf (vop-parse-inherits parse) inherits)
 
     (parse-define-vop parse specs)
@@ -1604,79 +1684,144 @@
                          (vop-parse-more-results parse)
                          (vop-parse-result-types parse)
                          nil)
-    `(progn
-       (eval-when (:compile-toplevel)
-         (setf (gethash ',name *backend-parsed-vops*) ',parse))
-       (register-vop-parse
-        ,@(macrolet
-              ((quotify-slots ()
-                 (collect ((forms))
-                   (dolist (x vop-parse-slot-names (cons 'list (forms)))
-                     (let ((reader (package-symbolicate (sb-xc:symbol-package 'vop-parse)
-                                                        "VOP-PARSE-" x)))
-                       (forms
-                        (case x
-                          (source-location ''(source-location))
-                          ((temps args results) `(quotify-list (,reader parse)))
-                          ((more-args more-results) `(quotify (,reader parse)))
-                          (t `(list 'quote (,reader parse))))))))))
-            (labels ((quotify (operand-or-nil)
-                       (when operand-or-nil
-                         (list 'quote (quotify-1 operand-or-nil))))
-                     (quotify-list (operands)
-                       (list 'quote (mapcar #'quotify-1 operands)))
-                     (quotify-1 (x) ; Return everything except the KIND, quoted
-                       `(,(operand-parse-name x)
-                         ,(operand-parse-target x) ,(operand-parse-temp x)
-                         ,(operand-parse-born x) ,(operand-parse-dies x)
-                         ,(operand-parse-load-tn x) ,(operand-parse-load x)
-                         ,(operand-parse-scs x) ,(operand-parse-offset x))))
-              (quotify-slots))))
-       ,@(unless (eq (vop-parse-body parse) :unspecified)
-           `((let ((,n-res ,(set-up-vop-info inherited-parse parse)))
-               (store-vop-info ,n-res)
-               ,@(set-up-fun-translation parse n-res))))
-       ',name)))
+    (if set
+        `(progn
+           (eval-when (:compile-toplevel)
+             (setf (gethash ',name *backend-parsed-vops*) ',parse))
+           (register-vop-parse
+            ,@(macrolet
+                  ((quotify-slots ()
+                     (collect ((forms))
+                       (dolist (x vop-parse-slot-names (cons 'list (forms)))
+                         (let ((reader (package-symbolicate (sb-xc:symbol-package 'vop-parse)
+                                                            "VOP-PARSE-" x)))
+                           (forms
+                            (case x
+                              (source-location ''(source-location))
+                              ((temps args results) `(quotify-list (,reader parse)))
+                              ((more-args more-results) `(quotify (,reader parse)))
+                              (t `(list 'quote (,reader parse))))))))))
+                (labels ((quotify (operand-or-nil)
+                           (when operand-or-nil
+                             (list 'quote (quotify-1 operand-or-nil))))
+                         (quotify-list (operands)
+                           (list 'quote (mapcar #'quotify-1 operands)))
+                         (quotify-1 (x) ; Return everything except the KIND, quoted
+                           `(,(operand-parse-name x)
+                             ,(operand-parse-target x) ,(operand-parse-temp x)
+                             ,(operand-parse-born x) ,(operand-parse-dies x)
+                             ,(operand-parse-load-tn x) ,(operand-parse-load x)
+                             ,(operand-parse-scs x) ,(operand-parse-offset x))))
+                  (quotify-slots))))
+           ,@(unless (eq (vop-parse-body parse) :unspecified)
+               `((let ((,n-res ,(set-up-vop-info inherited-parse parse)))
+                   (store-vop-info ,n-res)
+                   ,@(set-up-fun-translation parse n-res))))
+           ',name)
+        `(let ((info ,(set-up-vop-info inherited-parse parse)))
+           (setf (vop-info-type info)
+                 (specifier-type (template-type-specifier info)))
+           info))))
+
+;;; (inline-vop
+;;;     (((param unsigned-reg unsigned-num :to :save) arg)
+;;;      ((temp unsigned-reg unsigned-num))
+;;;      ((temp2))) ;; will reuse the previous specifications
+;;;     ((result unsigned-reg unsigned-num))
+;;;   (inst x result temp param))
+(defmacro inline-vop (vars results &body body)
+  (collect ((input)
+            (args)
+            (arg-types)
+            (infos)
+            (temps)
+            (results)
+            (result-types))
+    (loop for (var arg) in vars
+          for (name this-sc) = var
+          for (nil sc type . rest) = (if this-sc
+                                         var
+                                         prev)
+          for prev = (if this-sc
+                         var
+                         prev)
+          do (cond ((eq name :info)
+                    (infos this-sc)
+                    (input arg))
+                   (arg
+                    (input arg)
+                    (args (list* name :scs (list sc) rest))
+                    (arg-types (or type '*)))
+                   (t
+                    (temps `(:temporary (:sc ,sc ,@rest)
+                                        ,name)))))
+    (loop for result in results
+          for (name this-sc) = result
+          for (nil sc type . rest) = (if this-sc
+                                         result
+                                         prev)
+          for prev = (if this-sc
+                         result
+                         prev)
+          do (results (list* name :scs (list sc) rest))
+             (result-types (or type '*)))
+    `(inline-%primitive
+      ,(eval (%define-vop nil nil
+                          (delete nil
+                                  (list* (and (args)
+                                              (list* :args (args)))
+                                         (and (arg-types)
+                                              (list* :arg-types (arg-types)))
+                                         (and (results)
+                                              (list* :results (results)))
+                                         (and (result-types)
+                                              (list* :result-types (result-types)))
+                                         (and (infos)
+                                              (list* :info (infos)))
+                                         (list* :generator 0 body)
+                                         (temps)))
+                          nil))
+      ,@(input))))
 
 (macrolet
-   ((def ()
-     `(defun register-vop-parse ,vop-parse-slot-names
-       ;; Try to share each OPERAND-PARSE structure with a similar existing one.
-       (labels ((share-list (operand-specs accessor kind)
-                  (let ((new (mapcar (lambda (x) (share x kind)) operand-specs)))
-                    (dohash ((key parse) *backend-parsed-vops* :result new)
-                      (declare (ignore key))
-                      (when (equal (funcall accessor parse) new)
-                        (return (funcall accessor parse))))))
-                (share (operand-spec kind)
-                  ;; OPERAND-PARSE structures are immutable. Scan all vops for one
-                  ;; with an operand matching OPERAND-SPEC, and use that if found.
-                  (destructuring-bind (name targ temp born dies load-tn load scs offs)
-                      operand-spec
-                    (let ((op (make-operand-parse
-                               :name name :kind kind :target targ :temp temp
-                               :born born :dies dies :load-tn load-tn :load load
-                               :scs scs :offset offs)))
-                      (dohash ((key parse) *backend-parsed-vops* :result op)
-                        (declare (ignore key))
-                        (awhen (find op (vop-parse-operands parse) :test #'operand=)
-                          (return it))))))
-                (operand= (a b)
-                  ;; EQUALP is too weak a comparator for arbitrary sexprs,
-                  ;; since (EQUALP "foo" #(#\F #\O #\O)) is T, not that
-                  ;; we expect such weirdness in the LOAD-IF expression.
-                  (and (equal (operand-parse-load a) (operand-parse-load b))
-                       (equalp a b))))
-         (setq temps (share-list temps #'vop-parse-temps :temporary)
-               args (share-list args #'vop-parse-args :argument)
-               results (share-list results #'vop-parse-results :result))
-         (when more-args (setq more-args (share more-args :more-argument)))
-         (when more-results (setq more-results (share more-results :more-result))))
-       (let ((parse
-              (make-vop-parse ,@(mapcan (lambda (x) (list (keywordicate x) x))
-                                        vop-parse-slot-names))))
-         (set-vop-parse-operands parse)
-         (setf (gethash name *backend-parsed-vops*) parse)))))
+    ((def ()
+       `(defun register-vop-parse ,vop-parse-slot-names
+          ;; Try to share each OPERAND-PARSE structure with a similar existing one.
+          (labels ((share-list (operand-specs accessor kind)
+                     (let ((new (mapcar (lambda (x) (share x kind)) operand-specs)))
+                       (dohash ((key parse) *backend-parsed-vops* :result new)
+                         (declare (ignore key))
+                         (when (equal (funcall accessor parse) new)
+                           (return (funcall accessor parse))))))
+                   (share (operand-spec kind)
+                     ;; OPERAND-PARSE structures are immutable. Scan all vops for one
+                     ;; with an operand matching OPERAND-SPEC, and use that if found.
+                     (destructuring-bind (name targ temp born dies load-tn load scs offs)
+                         operand-spec
+                       (let ((op (make-operand-parse
+                                  :name name :kind kind :target targ :temp temp
+                                  :born born :dies dies :load-tn load-tn :load load
+                                  :scs scs :offset offs)))
+                         (dohash ((key parse) *backend-parsed-vops* :result op)
+                           (declare (ignore key))
+                           (awhen (find op (vop-parse-operands parse) :test #'operand=)
+                             (return it))))))
+                   (operand= (a b)
+                     ;; EQUALP is too weak a comparator for arbitrary sexprs,
+                     ;; since (EQUALP "foo" #(#\F #\O #\O)) is T, not that
+                     ;; we expect such weirdness in the LOAD-IF expression.
+                     (and (equal (operand-parse-load a) (operand-parse-load b))
+                          (equalp a b))))
+            (setq temps (share-list temps #'vop-parse-temps :temporary)
+                  args (share-list args #'vop-parse-args :argument)
+                  results (share-list results #'vop-parse-results :result))
+            (when more-args (setq more-args (share more-args :more-argument)))
+            (when more-results (setq more-results (share more-results :more-result))))
+          (let ((parse
+                  (make-vop-parse ,@(mapcan (lambda (x) (list (keywordicate x) x))
+                                            vop-parse-slot-names))))
+            (set-vop-parse-operands parse)
+            (setf (gethash name *backend-parsed-vops*) parse)))))
   (def))
 
 (defun store-vop-info (vop-info)
@@ -1815,8 +1960,8 @@
          (result-count (length (vop-parse-results parse)))
          (info-count (length (vop-parse-info-args parse)))
          (noperands (+ arg-count result-count info-count))
-         (n-node (sb-xc:gensym))
-         (n-block (sb-xc:gensym))
+         (n-node (gensym))
+         (n-block (gensym))
          (n-template (gensym)))
 
     (when (or (vop-parse-more-args parse) (vop-parse-more-results parse))
@@ -1833,7 +1978,7 @@
         (collect ((ibinds)
                   (ivars))
           (dolist (info (subseq operands arg-count (+ arg-count info-count)))
-            (let ((temp (sb-xc:gensym)))
+            (let ((temp (gensym)))
               (ibinds `(,temp ,info))
               (ivars temp)))
 
@@ -1875,7 +2020,7 @@
          (fixed-args (butlast args))
          (fixed-results (butlast results))
          (n-node (gensym))
-         (n-block (sb-xc:gensym))
+         (n-block (gensym))
          (n-template (gensym)))
 
     (unless (or (vop-parse-more-args parse)
@@ -1990,17 +2135,18 @@
                    (when (and ,tn-var (not (eq ,tn-var :more)))
                      (,bod ,tn-var)))))))))))
 
-;;; Iterate over all the IR2 blocks in PHYSENV, in emit order.
-(defmacro do-physenv-ir2-blocks ((block-var physenv &optional result)
-                                 &body body)
-  (once-only ((n-physenv physenv))
-    (once-only ((n-first `(lambda-block (physenv-lambda ,n-physenv))))
+;;; Iterate over all the IR2 blocks in the environment ENV, in emit
+;;; order.
+(defmacro do-environment-ir2-blocks ((block-var env &optional result)
+                                     &body body)
+  (once-only ((n-env env))
+    (once-only ((n-first `(lambda-block (environment-lambda ,n-env))))
       (once-only ((n-tail `(block-info
                             (component-tail
                              (block-component ,n-first)))))
         `(do ((,block-var (block-info ,n-first)
                           (ir2-block-next ,block-var)))
              ((or (eq ,block-var ,n-tail)
-                  (not (eq (ir2-block-physenv ,block-var) ,n-physenv)))
+                  (not (eq (ir2-block-environment ,block-var) ,n-env)))
               ,result)
            ,@body)))))
