@@ -510,7 +510,7 @@
             '((old-fp :target old-fp-pass)
               (return-pc :target return-pc-pass)))
         ,@(unless variable
-            '((args :more t :scs (descriptor-reg)))))
+            '((args :more t :scs (descriptor-reg control-stack)))))
        ,@(when (eq return :fixed)
            '((:results (values :more t))))
        (:save-p ,(if (eq return :tail) :compute-only t))
