@@ -39,6 +39,7 @@
   (:results)
   (:vop-var vop)
   (:node-var node)
+  (:args-var args)
   (:temporary (:sc unsigned-reg) val-temp)
   (:generator 1
     (cond #+ubsan
@@ -50,10 +51,19 @@
            (inst mov :dword (vector-len-ea object)
                  (or (encode-value-if-immediate value) value)))
           (t
-           ;; gencgc does not need to emit the barrier for constructors
-           (unless (or (eq name :allocator)
-                       (sb-c::set-slot-old-p node))
-             (emit-gc-store-barrier object nil val-temp (vop-nth-arg 1 vop) value))
+           (let* ((value-tn (tn-ref-tn (tn-ref-across args)))
+                  (prim-type (sb-c::tn-primitive-type value-tn))
+                  (scs (and prim-type
+                            (sb-c::primitive-type-scs prim-type))))
+             (unless (and
+                      ;; Can this TN be boxed after the allocator?
+                      (or (singleton-p scs)
+                          (not (member descriptor-reg-sc-number scs)))
+                      (or
+                       ;; gencgc does not need to emit the barrier for constructors
+                       (eq name :allocator)
+                       (sb-c::set-slot-old-p node)))
+               (emit-gc-store-barrier object nil val-temp (vop-nth-arg 1 vop) value)))
            (gen-cell-set (object-slot-ea object offset lowtag) value val-temp)))))
 
 (define-vop (compare-and-swap-slot)
@@ -164,7 +174,7 @@
                                            (symbol-tls-index-ea symbol)
                                            (tls-index-of symbol)))
                  (inst add cell thread-tn)))
-          (inst cmp :dword (ea cell) no-tls-value-marker-widetag)
+          (inst cmp :qword (ea cell) no-tls-value-marker)
           (inst jmp :ne STORE)
           (emit-symbol-write-barrier symbol nil val-temp (vop-nth-arg 1 vop) value)
           (get-symbol-value-slot-ea cell symbol)
@@ -262,7 +272,7 @@
                                 (symbol-tls-index-ea symbol)
                                 (tls-index-of symbol)))
       (inst add cell thread-tn)
-      (inst cmp :dword (ea cell) no-tls-value-marker-widetag)
+      (inst cmp :qword (ea cell) no-tls-value-marker)
       (inst jmp :ne CAS))
       ;; GLOBAL. All logic that follows is for both + and - sb-thread
       (emit-symbol-write-barrier symbol nil cell (vop-nth-arg 2 vop) new)
@@ -323,7 +333,7 @@
               (setq symbol-reg symbol)))
 
            ;; Load the global value if the TLS value didn't exist
-           (inst cmp :dword value no-tls-value-marker-widetag)
+           (inst cmp :qword value no-tls-value-marker)
            (inst cmov :e value
                  (if (and known-symbol-p (sc-is symbol immediate))
                      (symbol-slot-ea known-symbol symbol-value-slot) ; MOV Rxx, imm32
@@ -388,8 +398,8 @@
       (:temporary (:sc unsigned-reg) temp)
       (:generator 9
         (inst mov :dword temp (tls-index-of object))
-        (inst mov :dword temp (thread-tls-ea temp))
-        (inst cmp :dword temp no-tls-value-marker-widetag)
+        (inst mov :qword temp (thread-tls-ea temp))
+        (inst cmp :qword temp no-tls-value-marker)
         (inst cmov :dword :e temp (symbol-value-slot-ea object))
         (inst cmp :byte temp unbound-marker-widetag))))
 
@@ -482,15 +492,16 @@
     (inst mov res (ea n-word-bytes base-ptr)) ; 1 word beyond the header
     (inst and res (lognot fixnum-tag-mask))))
 
+(aver (= sb-impl::package-id-bits 16))
 (define-vop ()
   (:args (symbol :scs (descriptor-reg)))
   (:results (result :scs (unsigned-reg)))
   (:result-types positive-fixnum)
-  (:translate sb-impl::symbol-package-id)
+  (:translate symbol-package-id)
   (:policy :fast-safe)
-  (:generator 2 ; ASSUMPTION: symbol-package-bits = 16
-   (inst mov :word result (object-slot-ea symbol symbol-name-slot (- other-pointer-lowtag 6)))
-   (inst movzx '(:word :dword) result result)))
+  (:generator 2
+   (inst movzx '(:word :dword) result
+         (object-slot-ea symbol symbol-name-slot (- other-pointer-lowtag 6)))))
 (define-vop ()
   (:args (symbol :scs (descriptor-reg)))
   (:results (result :scs (descriptor-reg) :from :load))
@@ -502,6 +513,8 @@
 
 ;;;; fdefinition (FDEFN) objects
 
+;;; IR2-CONVERT-GLOBAL-VAR makes direct use of this vop by name,
+;;; so even though it has a :ref-trans in objdef, we also need the vop.
 (define-vop (fdefn-fun cell-ref)        ; /pfw - alpha
   (:variant fdefn-fun-slot other-pointer-lowtag))
 
@@ -521,10 +534,9 @@
       (inst jmp :e err-lab))
     RETRY))
 
-#-immobile-code
 (define-vop (set-fdefn-fun)
   (:policy :fast-safe)
-  (:translate (setf fdefn-fun))
+  #-immobile-code (:translate (setf fdefn-fun))
   (:args (function :scs (descriptor-reg) :target result)
          (fdefn :scs (descriptor-reg)))
   (:temporary (:sc unsigned-reg) raw)
@@ -541,7 +553,7 @@
     (move result function)))
 #+immobile-code
 (progn
-(define-vop (set-fdefn-fun)
+(define-vop (set-direct-callable-fdefn-fun)
   (:args (fdefn :scs (descriptor-reg))
          (function :scs (descriptor-reg))
          (raw-word :scs (unsigned-reg)))
@@ -569,6 +581,7 @@
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
   (:args (fdefn :scs (descriptor-reg)))
+  (:temporary (:sc unsigned-reg) temp)
   (:vop-var vop)
   (:generator 38
     ;; Change the JMP instruction to INT3 so that a trap occurs in the fdefn
@@ -584,8 +597,9 @@
     ;; and resume at undefined-tramp. However, CALL-SYMBOL jumps via raw-addr if
     ;; its callable object was not a function. In that case RAX holds a symbol,
     ;; so we're OK because we can identify the undefined function.
-    (storew (make-fixup 'undefined-tramp :assembly-routine)
-            fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
+    ;; Technically this could be computed using "LEA temp, [RIP-disp]"
+    (inst mov temp (ea (make-fixup 'undefined-tramp :assembly-routine*)))
+    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 ;;;; binding and unbinding
 
@@ -754,15 +768,23 @@
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg)))
   (:info offset)
+  (:vop-var vop)
+  ;; temp is wasted if we don't need a barrier, which we almost never do
+  (:temporary (:sc unsigned-reg) temp)
   (:generator 4
-    ;; TODO: gencgc does not need EMIT-GC-STORE-BARRIER here, but other other GC strategies might.
+    (let* ((value-tn (tn-ref-tn (tn-ref-across (vop-args vop))))
+           (prim-type (sb-c::tn-primitive-type value-tn))
+           (scs (and prim-type (sb-c::primitive-type-scs prim-type))))
+      (when (and (not (singleton-p scs))
+                 (member descriptor-reg-sc-number scs))
+        (emit-gc-store-barrier object nil temp (vop-nth-arg 1 vop) value)))
     (storew value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
 (define-vop (closure-init-from-fp)
   (:args (object :scs (descriptor-reg)))
   (:info offset)
   (:generator 4
-    ;; TODO: gencgc does not need EMIT-GC-STORE-BARRIER here, but other other GC strategies might.
+    ;; RBP-TN looks like a fixnum (non-pointer) so no barrier
     (storew rbp-tn object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
 ;;;; value cell hackery
