@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 (defun ea-for-sse-stack (tn &optional (base rbp-tn))
   (ea (frame-byte-offset (1+ (tn-offset tn))) base))
@@ -80,36 +80,44 @@
   (int-sse-reg single-sse-reg double-sse-reg)
   (int-sse-reg single-sse-reg double-sse-reg))
 
-(define-vop (move-from-sse)
-  (:args (x :scs (single-sse-reg double-sse-reg int-sse-reg)))
-  (:results (y :scs (descriptor-reg)))
-  (:node-var node)
-  (:note "SSE to pointer coercion")
-  (:generator 13
-     (fixed-alloc y simd-pack-widetag simd-pack-size node)
-       ;; see *simd-pack-element-types*
-     (storew (fixnumize
-              (sc-case x
-                (single-sse-reg 1)
-                (double-sse-reg 2)
-                (int-sse-reg 0)
-                (t 0)))
-         y simd-pack-tag-slot other-pointer-lowtag)
-     (let ((ea (make-ea-for-object-slot
-                y simd-pack-lo-value-slot other-pointer-lowtag)))
-       (if (float-sse-p x)
-           (inst movaps ea x)
-           (inst movdqa ea x)))))
-(define-move-vop move-from-sse :move
-  (int-sse-reg single-sse-reg double-sse-reg) (descriptor-reg))
+(macrolet ((define-move-from-sse (type tag &rest scs)
+             (let ((name (symbolicate "MOVE-FROM-SSE/" type)))
+               `(progn
+                  (define-vop (,name)
+                    (:args (x :scs ,scs))
+                    (:results (y :scs (descriptor-reg)))
+                    #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
+                    (:node-var node)
+                    (:arg-types ,type)
+                    (:note "AVX2 to pointer coercion")
+                    (:generator 13
+                      (alloc-other simd-pack-widetag simd-pack-size y node nil thread-tn)
+                      (storew (fixnumize ,tag)
+                              y simd-pack-tag-slot other-pointer-lowtag)
+                      (let ((ea (object-slot-ea y simd-pack-lo-value-slot other-pointer-lowtag)))
+                        (if (float-sse-p x)
+                            (inst movaps ea x)
+                            (inst movdqa ea x)))))
+                  (define-move-vop ,name :move
+                    ,scs (descriptor-reg))))))
+  ;; see +simd-pack-element-types+
+  (define-move-from-sse simd-pack-single 0 single-sse-reg)
+  (define-move-from-sse simd-pack-double 1 double-sse-reg)
+  (define-move-from-sse simd-pack-ub8 2 int-sse-reg)
+  (define-move-from-sse simd-pack-ub16 3 int-sse-reg)
+  (define-move-from-sse simd-pack-ub32 4 int-sse-reg)
+  (define-move-from-sse simd-pack-ub64 5 int-sse-reg)
+  (define-move-from-sse simd-pack-sb8 6 int-sse-reg)
+  (define-move-from-sse simd-pack-sb16 7 int-sse-reg)
+  (define-move-from-sse simd-pack-sb32 8 int-sse-reg)
+  (define-move-from-sse simd-pack-sb64 9 int-sse-reg))
 
 (define-vop (move-to-sse)
   (:args (x :scs (descriptor-reg)))
   (:results (y :scs (int-sse-reg double-sse-reg single-sse-reg)))
   (:note "pointer to SSE coercion")
   (:generator 2
-    (let ((ea (make-ea-for-object-slot
-               x simd-pack-lo-value-slot other-pointer-lowtag)))
+    (let ((ea (object-slot-ea x simd-pack-lo-value-slot other-pointer-lowtag)))
       (if (float-sse-p y)
           (inst movaps y ea)
           (inst movdqa y ea)))))
@@ -187,10 +195,11 @@
   (:arg-types tagged-num unsigned-num unsigned-num)
   (:results (dst :scs (descriptor-reg) :from :load))
   (:result-types t)
+  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
   (:node-var node)
   (:generator 13
-    (fixed-alloc dst simd-pack-widetag simd-pack-size node)
-      ;; see *simd-pack-element-types*
+    (alloc-other simd-pack-widetag simd-pack-size dst node nil thread-tn)
+      ;; see +simd-pack-element-types+
     (storew tag dst simd-pack-tag-slot other-pointer-lowtag)
     (storew lo dst simd-pack-lo-value-slot other-pointer-lowtag)
     (storew hi dst simd-pack-hi-value-slot other-pointer-lowtag)))
@@ -203,36 +212,96 @@
   (:arg-types unsigned-num unsigned-num)
   (:temporary (:sc int-sse-reg) tmp)
   (:results (dst :scs (int-sse-reg)))
-  (:result-types simd-pack-int)
+  (:result-types simd-pack-ub64)
   (:generator 5
     (inst movq dst lo)
     (inst movq tmp hi)
     (inst punpcklqdq dst tmp)))
 
+(defmacro simd-pack-dispatch (pack &body body)
+  (check-type pack symbol)
+  `(let ((,pack ,pack))
+     (etypecase ,pack
+       ,@(map 'list (lambda (eltype)
+                   `((simd-pack ,eltype) ,@body))
+          +simd-pack-element-types+))))
+
 #-sb-xc-host
-(declaim (inline %make-simd-pack-ub32))
+(macrolet ((unpack-unsigned (pack bits)
+             `(simd-pack-dispatch ,pack
+                (let ((lo (%simd-pack-low ,pack))
+                      (hi (%simd-pack-high ,pack)))
+                  (values
+                   ,@(loop for pos by bits below 64 collect
+                           `(unpack-unsigned-1 ,bits ,pos lo))
+                   ,@(loop for pos by bits below 64 collect
+                           `(unpack-unsigned-1 ,bits ,pos hi))))))
+           (unpack-unsigned-1 (bits position ub64)
+             `(ldb (byte ,bits ,position) ,ub64)))
+  (declaim (inline %simd-pack-ub8s))
+  (defun %simd-pack-ub8s (pack)
+    (declare (type simd-pack pack))
+    (unpack-unsigned pack 8))
+
+  (declaim (inline %simd-pack-ub16s))
+  (defun %simd-pack-ub16s (pack)
+    (declare (type simd-pack pack))
+    (unpack-unsigned pack 16))
+
+  (declaim (inline %simd-pack-ub32s))
+  (defun %simd-pack-ub32s (pack)
+    (declare (type simd-pack pack))
+    (unpack-unsigned pack 32))
+
+  (declaim (inline %simd-pack-ub64s))
+  (defun %simd-pack-ub64s (pack)
+    (declare (type simd-pack pack))
+    (unpack-unsigned pack 64)))
+
 #-sb-xc-host
-(defun %make-simd-pack-ub32 (w x y z)
-  (declare (type (unsigned-byte 32) w x y z))
-  (%make-simd-pack-ub64 (logior w (ash x 32))
-                        (logior y (ash z 32))))
+(macrolet ((unpack-signed (pack bits)
+             `(simd-pack-dispatch ,pack
+                (let ((lo (%simd-pack-low ,pack))
+                      (hi (%simd-pack-high ,pack)))
+                  (values
+                   ,@(loop for pos by bits below 64 collect
+                           `(unpack-signed-1 ,bits ,pos lo))
+                   ,@(loop for pos by bits below 64 collect
+                           `(unpack-signed-1 ,bits ,pos hi))))))
+           (unpack-signed-1 (bits position ub64)
+             `(- (mod (+ (ldb (byte ,bits ,position) ,ub64)
+                         ,(expt 2 (1- bits)))
+                      ,(expt 2 bits))
+                 ,(expt 2 (1- bits)))))
+  (declaim (inline %simd-pack-sb8s))
+  (defun %simd-pack-sb8s (pack)
+    (declare (type simd-pack pack))
+    (unpack-signed pack 8))
+
+  (declaim (inline %simd-pack-sb16s))
+  (defun %simd-pack-sb16s (pack)
+    (declare (type simd-pack pack))
+    (unpack-signed pack 16))
+
+  (declaim (inline %simd-pack-sb32s))
+  (defun %simd-pack-sb32s (pack)
+    (declare (type simd-pack pack))
+    (unpack-signed pack 32))
+
+  (declaim (inline %simd-pack-sb64s))
+  (defun %simd-pack-sb64s (pack)
+    (declare (type simd-pack pack))
+    (unpack-signed pack 64)))
 
 #-sb-xc-host
 (progn
-  (declaim (inline %simd-pack-ub32s %simd-pack-ub64s))
-  (defun %simd-pack-ub32s (pack)
-    (declare (type simd-pack pack))
-    (let ((lo (%simd-pack-low pack))
-          (hi (%simd-pack-high pack)))
-      (values (ldb (byte 32 0) lo)
-              (ash lo -32)
-              (ldb (byte 32 0) hi)
-              (ash hi -32))))
-
-  (defun %simd-pack-ub64s (pack)
-    (declare (type simd-pack pack))
-    (values (%simd-pack-low pack)
-            (%simd-pack-high pack))))
+  (declaim (inline %make-simd-pack-ub32))
+  (defun %make-simd-pack-ub32 (w x y z)
+    (declare (type (unsigned-byte 32) w x y z))
+    (%make-simd-pack
+     #.(position '(unsigned-byte 32) +simd-pack-element-types+ :test #'equal)
+     (logior w (ash x 32))
+     (logior y (ash z 32)))))
 
 (define-vop (%make-simd-pack-double)
   (:translate %make-simd-pack-double)
@@ -292,14 +361,15 @@
                      (inst movss dst tmp)))))
 
 #-sb-xc-host
+(progn
 (declaim (inline %simd-pack-singles))
-#-sb-xc-host
 (defun %simd-pack-singles (pack)
   (declare (type simd-pack pack))
-  (values (%simd-pack-single-item pack 0)
-          (%simd-pack-single-item pack 1)
-          (%simd-pack-single-item pack 2)
-          (%simd-pack-single-item pack 3)))
+  (simd-pack-dispatch pack
+    (values (%simd-pack-single-item pack 0)
+            (%simd-pack-single-item pack 1)
+            (%simd-pack-single-item pack 2)
+            (%simd-pack-single-item pack 3)))))
 
 (defknown %simd-pack-double-item
   (simd-pack (integer 0 1)) double-float (flushable))
@@ -327,9 +397,10 @@
            (inst movsd dst tmp)))))
 
 #-sb-xc-host
+(progn
 (declaim (inline %simd-pack-doubles))
-#-sb-xc-host
 (defun %simd-pack-doubles (pack)
   (declare (type simd-pack pack))
-  (values (%simd-pack-double-item pack 0)
-          (%simd-pack-double-item pack 1)))
+  (simd-pack-dispatch pack
+    (values (%simd-pack-double-item pack 0)
+            (%simd-pack-double-item pack 1)))))
