@@ -116,6 +116,9 @@
 (defun emit-one-class-writer (class-slot-p)
   (emit-reader/writer :writer 1 class-slot-p))
 
+(defun emit-one-class-makunbound (class-slot-p)
+  (emit-reader/writer :makunbound 1 class-slot-p))
+
 (defun emit-two-class-reader (class-slot-p)
   (emit-reader/writer :reader 2 class-slot-p))
 
@@ -125,6 +128,9 @@
 (defun emit-two-class-writer (class-slot-p)
   (emit-reader/writer :writer 2 class-slot-p))
 
+(defun emit-two-class-makunbound (class-slot-p)
+  (emit-reader/writer :makunbound 2 class-slot-p))
+
 ;;; --------------------------------
 
 (defun emit-one-index-readers (class-slot-p)
@@ -136,6 +142,9 @@
 (defun emit-one-index-writers (class-slot-p)
   (emit-one-or-n-index-reader/writer :writer nil class-slot-p))
 
+(defun emit-one-index-makunbounds (class-slot-p)
+  (emit-one-or-n-index-reader/writer :makunbound nil class-slot-p))
+
 (defun emit-n-n-readers ()
   (emit-one-or-n-index-reader/writer :reader t nil))
 
@@ -145,6 +154,9 @@
 (defun emit-n-n-writers ()
   (emit-one-or-n-index-reader/writer :writer t nil))
 
+(defun emit-n-n-makunbounds ()
+  (emit-one-or-n-index-reader/writer :makunbound t nil))
+
 ;;; --------------------------------
 
 (defun emit-checking (metatypes applyp)
@@ -205,7 +217,7 @@
         (closure-variables ())
         (read-form (emit-slot-read-form class-slot-p 'index 'slots)))
     (ecase reader/writer
-      ((:reader :boundp)
+      ((:reader :boundp :makunbound)
        (setq instance (dfun-arg-symbol 0)
              arglist  (list instance)))
       (:writer (setq instance (dfun-arg-symbol 1)
@@ -242,6 +254,9 @@
                 (:boundp
                  `((let ((value ,read-form))
                      (return-from access (not (unbound-marker-p value))))))
+                (:makunbound
+                 `(progn (setf ,read-form +slot-unbound+)
+                         ,instance))
                 (:writer
                  `((return-from access (setf ,read-form ,(car arglist)))))))
           (funcall miss-fn ,@arglist))))))
@@ -263,6 +278,7 @@
     (ecase reader/writer
       (:reader (emit-boundp-check read-form miss-fn arglist))
       (:boundp `(not (unbound-marker-p ,read-form)))
+      (:makunbound `(progn (setf ,read-form +slot-unbound+) ,(car arglist)))
       (:writer `(setf ,read-form ,(car arglist))))))
 
 (defmacro emit-reader/writer-macro (reader/writer 1-or-2-class class-slot-p)
@@ -278,7 +294,7 @@
                                           class-slot-p)
   (multiple-value-bind (arglist metatypes)
       (ecase reader/writer
-        ((:reader :boundp)
+        ((:reader :boundp :makunbound)
          (values (list (dfun-arg-symbol 0))
                  '(standard-instance)))
         (:writer (values (list (dfun-arg-symbol 0) (dfun-arg-symbol 1))
