@@ -103,13 +103,11 @@ int n_logevents;
 #define OS_SA_NODEFER SA_NODEFER
 #endif
 
-static inline void
-sigcopyset(sigset_t *new, sigset_t *old)
-{
+static inline void sigcopyset(sigset_t *to, sigset_t *from) {
 #ifdef ADDRESS_SANITIZER
-    sigemptyset(new);
+    sigemptyset(to);
 #endif
-    memcpy(new, old, REAL_SIGSET_SIZE_BYTES);
+    memcpy(to, from, REAL_SIGSET_SIZE_BYTES);
 }
 
 /* When we catch an internal error, should we pass it back to Lisp to
@@ -548,7 +546,9 @@ check_deferrables_blocked_or_lose(sigset_t *sigset)
         lose("deferrables unblocked");
 }
 
+#ifdef LISP_FEATURE_RISCV
 int sigaction_does_not_mask;
+#endif
 static void assert_blockables_blocked()
 {
 #ifdef LISP_FEATURE_RISCV
@@ -656,7 +656,7 @@ unblock_deferrable_signals(sigset_t *where)
 #ifndef LISP_FEATURE_SB_SAFEPOINT
 // This function previously had an #ifdef guard precluding doing anything for
 // win32, which was redundant because SB_SAFEPOINT is always defined for win32.
-void unblock_gc_signals(void) {
+void unblock_gc_stop_signal(void) {
     thread_sigmask(SIG_UNBLOCK, &gc_sigset, 0);
 }
 #endif
@@ -690,16 +690,18 @@ check_interrupts_enabled_or_lose(os_context_t *context)
         lose ("in pseudo atomic section");
 }
 
+/* Note that the comment from rev aa0ed5a420 seems back-ass-wards.
+ * "if there is no pending signal .. because that means deferrable are blocked"?
+ * How would NO pending signal imply that deferrables are blocked? */
+
 /* Save sigset (or the current sigmask if 0) if there is no pending
  * handler, because that means that deferabbles are already blocked.
  * The purpose is to avoid losing the pending gc signal if a
  * deferrable interrupt async unwinds between clearing the pseudo
  * atomic and trapping to GC.*/
 #ifndef LISP_FEATURE_SB_SAFEPOINT
-void
-maybe_save_gc_mask_and_block_deferrables(sigset_t *sigset)
+void maybe_save_gc_mask_and_block_deferrables(os_context_t *context)
 {
-#ifndef LISP_FEATURE_WIN32
     struct thread *thread = get_sb_vm_thread();
     struct interrupt_data *data = &thread_interrupt_data(thread);
     sigset_t oldset;
@@ -715,8 +717,13 @@ maybe_save_gc_mask_and_block_deferrables(sigset_t *sigset)
     if ((!data->pending_handler) &&
         (!data->gc_blocked_deferrables)) {
         data->gc_blocked_deferrables = 1;
-        if (sigset) {
-            /* This is the sigmask of some context. */
+        if (context) {
+            /* When there is a sigcontext, do the following:
+             * - 'pending_mask' stores the mask from just prior to receipt of the signal
+             * - On return to the interrupt point, the signal mask is restored
+             *   to what it was PLUS all deferrable signals
+             * - On return from this function, no change to the current mask */
+            sigset_t *sigset = os_context_sigmask_addr(context);
             sigcopyset(&data->pending_mask, sigset);
             sigaddset_deferrable(sigset);
             thread_sigmask(SIG_SETMASK,&oldset,0);
@@ -724,14 +731,14 @@ maybe_save_gc_mask_and_block_deferrables(sigset_t *sigset)
         } else {
             /* Operating on the current sigmask. Save oldset and
              * unblock gc signals. In the end, this is equivalent to
-             * blocking the deferrables. */
+             * blocking the deferrables and also SIGPIPE and SIGPROF neither of which
+             * is usually deferred, for reasons cited in sigaddset_blockable */
             sigcopyset(&data->pending_mask, &oldset);
-            unblock_gc_signals();
+            unblock_gc_stop_signal();
             return;
         }
     }
     thread_sigmask(SIG_SETMASK,&oldset,0);
-#endif
 }
 #endif
 
@@ -765,7 +772,7 @@ check_interrupt_context_or_lose(os_context_t *context)
 #if defined(LISP_FEATURE_GENCGC) && !GENCGC_IS_PRECISE
     int interrupts_enabled = (read_TLS(INTERRUPTS_ENABLED,thread) != NIL);
     int gc_inhibit = (read_TLS(GC_INHIBIT,thread) != NIL);
-    int gc_pending = (read_TLS(GC_PENDING,thread) == T);
+    int gc_pending = (read_TLS(GC_PENDING,thread) == LISP_T);
     int pseudo_atomic_interrupted = get_pseudo_atomic_interrupted(thread);
     int in_race_p = in_leaving_without_gcing_race_p(thread);
     int safepoint_active = 0;
@@ -1035,7 +1042,7 @@ interrupt_internal_error(os_context_t *context, boolean continuable)
     // describe_internal_error(context); // uncomment me for debugging
 
     funcall2(StaticSymbolFunction(INTERNAL_ERROR), context_sap,
-             continuable ? T : NIL);
+             continuable ? LISP_T : NIL);
 
     undo_fake_foreign_function_call(context); /* blocks signals again */
     if (continuable)
@@ -1133,7 +1140,7 @@ interrupt_handle_pending(os_context_t *context)
          /* Test for T and not for != NIL since the value :IN-PROGRESS
           * used to be used in SUB-GC as part of the mechanism to
           * supress recursive gcs.*/
-        if (read_TLS(GC_PENDING,thread) == T) {
+        if (read_TLS(GC_PENDING,thread) == LISP_T) {
 
             /* Two reasons for doing this. First, if there is a
              * pending handler we don't want to run. Second, we are
@@ -1209,7 +1216,7 @@ interrupt_handle_pending(os_context_t *context)
         run_deferred_handler(data, context);
     }
 #ifdef LISP_FEATURE_SB_SAFEPOINT
-    if (read_TLS(THRUPTION_PENDING,thread)==T)
+    if (read_TLS(THRUPTION_PENDING,thread)==LISP_T)
         /* Special case for the following situation: There is a
          * thruption pending, but a signal had been deferred.  The
          * pitstop at the top of this function could only take care
@@ -1261,7 +1268,7 @@ interrupt_handle_now(int signal, siginfo_t *info, os_context_t *context)
          * allow signals again when it sees fit. */
         /* handler.lisp will hide from the GC, will be enabled in the handler itself.
          * Not a problem for the conservative GC. */
-        unblock_gc_signals();
+        unblock_gc_stop_signal();
 #endif
 
         WITH_GC_AT_SAFEPOINTS_ONLY()
@@ -1377,7 +1384,7 @@ can_handle_now(void *handler, struct interrupt_data *data,
         event3("can_handle_now(%p,%d): deferred (RACE=%d)", handler, signal,
                in_leaving_without_gcing_race_p(thread));
         store_signal_data_for_later(data,handler,signal,info,context);
-        write_TLS(INTERRUPT_PENDING, T,thread);
+        write_TLS(INTERRUPT_PENDING, LISP_T, thread);
         answer = 0;
     }
     /* a slightly confusing test. arch_pseudo_atomic_atomic() doesn't
@@ -1427,14 +1434,13 @@ sig_stop_for_gc_handler(int __attribute__((unused)) signal,
      * pseudo atomic until gc is finally allowed. */
     if (read_TLS(GC_INHIBIT,thread) != NIL) {
         event0("stop_for_gc deferred for *GC-INHIBIT*");
-        write_TLS(STOP_FOR_GC_PENDING,T,thread);
+        write_TLS(STOP_FOR_GC_PENDING, LISP_T, thread);
         return;
     } else if (arch_pseudo_atomic_atomic(context)) {
         event0("stop_for_gc deferred for PA");
-        write_TLS(STOP_FOR_GC_PENDING,T,thread);
+        write_TLS(STOP_FOR_GC_PENDING, LISP_T, thread);
         arch_set_pseudo_atomic_interrupted(context);
-        maybe_save_gc_mask_and_block_deferrables
-            (os_context_sigmask_addr(context));
+        maybe_save_gc_mask_and_block_deferrables(context);
         return;
     }
 
@@ -1560,7 +1566,6 @@ extern int *os_context_flags_addr(os_context_t *context);
 
 extern lispobj call_into_lisp(lispobj fun, lispobj *args, int nargs);
 extern void post_signal_tramp(void);
-extern void call_into_lisp_tramp(void);
 
 void
 arrange_return_to_c_function(os_context_t *context,
@@ -1623,7 +1628,7 @@ arrange_return_to_c_function(os_context_t *context,
 #if defined(LISP_FEATURE_DARWIN)
     uint32_t *register_save_area = (uint32_t *)os_allocate(0x40);
 
-    /* 1. os_validate (malloc/mmap) register_save_block
+    /* 1. allocate (malloc/mmap) register_save_block
      * 2. copy register state into register_save_block
      * 3. put a pointer to register_save_block in a register in the context
      * 4. set the context's EIP to point to a trampoline which:
@@ -1749,13 +1754,7 @@ arrange_return_to_c_function(os_context_t *context,
 void
 arrange_return_to_lisp_function(os_context_t *context, lispobj function)
 {
-#if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_X86)
-    arrange_return_to_c_function(context,
-                                 (call_into_lisp_lookalike)call_into_lisp_tramp,
-                                 function);
-#else
     arrange_return_to_c_function(context, call_into_lisp, function);
-#endif
 }
 
 // These have undefined_alien_function tramp in x-assem.S
@@ -1773,6 +1772,7 @@ undefined_alien_function(void)
 }
 #endif
 
+#ifndef LISP_FEATURE_WIN32
 void lower_thread_control_stack_guard_page(struct thread *th)
 {
     protect_control_stack_guard_page(0, th);
@@ -1789,12 +1789,14 @@ void reset_thread_control_stack_guard_page(struct thread *th)
     th->state_word.control_stack_guard_page_protected = 1;
     fprintf(stderr, "INFO: Control stack guard page reprotected\n");
 }
+#endif
 
 boolean
 handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
 {
     struct thread *th=get_sb_vm_thread();
 
+#ifndef LISP_FEATURE_WIN32
     if(addr >= CONTROL_STACK_HARD_GUARD_PAGE(th) &&
        addr < CONTROL_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
         fake_foreign_function_call(context);
@@ -1845,8 +1847,10 @@ handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
         reset_thread_control_stack_guard_page(th);
         return 1;
     }
-    else if(addr >= BINDING_STACK_HARD_GUARD_PAGE(th) &&
-            addr < BINDING_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
+    else
+#endif // !LISP_FEATURE_WIN32
+    if(addr >= BINDING_STACK_HARD_GUARD_PAGE(th) &&
+       addr < BINDING_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
         lose("Binding stack exhausted");
     }
     else if(addr >= BINDING_STACK_GUARD_PAGE(th) &&
